import re
from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import Field, root_validator

from .base import BaseCharacter, GenshinModel

__all__ = [
    "HyakuninIkkiCharacter",
    "HyakuninIkkiSkill",
    "HyakuninIkkiBattle",
    "HyakuninIkkiChallenge",
    "HyakuninIkki",
    "LabyrinthWarriorsCharacter",
    "LabyrinthWarriorsRune",
    "LabyrinthWarriorsChallenge",
    "LabyrinthWarriors",
    "ChineseActivity",
    "Activities",
]

# ---------------------------------------------------------
# Hyakunin Ikki:


class HyakuninIkkiCharacter(BaseCharacter):
    """A Hyakunin Ikki character"""

    level: int
    trial: bool = Field(galias="is_trail_avatar")


class HyakuninIkkiSkill(GenshinModel):
    """A Hyakunin Ikki skill"""

    id: int
    name: str
    icon: str
    description: str = Field(galias="desc")


class HyakuninIkkiBattle(GenshinModel):
    """A Hyakunin Ikki battle"""

    characters: List[HyakuninIkkiCharacter] = Field(galias="avatars")
    skills: List[HyakuninIkkiSkill] = Field(galias="skills")


class HyakuninIkkiChallenge(GenshinModel):
    """A Hyakunin Ikki challenge"""

    id: int = Field(galias="challenge_id")
    name: str = Field(galias="challenge_name")
    difficulty: int
    multiplier: int = Field(galias="score_multiple")
    score: int = Field(galias="max_score")
    medal_icon: str = Field(galias="heraldry_icon")

    battles: List[HyakuninIkkiBattle] = Field(galias="lineups")

    @property
    def medal(self) -> str:
        match = re.search(r"heraldry_(\w+)\.png", self.medal_icon)
        return match.group(1) if match else ""


class HyakuninIkki(GenshinModel):
    """A Hyakunin Ikki event"""

    challenges: List[HyakuninIkkiChallenge] = Field(galias="records")


# ---------------------------------------------------------
# Labyrinth Warriors:


class LabyrinthWarriorsCharacter(BaseCharacter):
    """A Labyrinth Warriors character"""

    level: int


class LabyrinthWarriorsRune(GenshinModel):
    """A Labyrinth Warriors rune"""

    id: int
    icon: str
    name: str
    description: str = Field(galias="desc")
    element: str


class LabyrinthWarriorsChallenge(GenshinModel):
    """A Labyrinth Warriors challenge"""

    id: int = Field(galias="challenge_id")
    name: str = Field(galias="challenge_name")
    passed: bool = Field(galias="is_passed")
    level: int = Field(galias="settled_level")

    main_characters: List[LabyrinthWarriorsCharacter] = Field(galias="main_avatars")
    support_characters: List[LabyrinthWarriorsCharacter] = Field(galias="support_avatars")
    runes: List[LabyrinthWarriorsRune]


class LabyrinthWarriors(GenshinModel):
    """A Labyrinth Warriors event"""

    challenges: List[LabyrinthWarriorsChallenge] = Field(galias="records")


# ---------------------------------------------------------
# Energy Amplifier Fruition


class EnergyAmplifierCharacter(BaseCharacter):
    """An energy amplifier character"""

    level: int


class EnergyAmplifierCriteria(GenshinModel):
    """An energy amplifier challenge criteria"""

    id: int
    description: str = Field(galias="desc")
    score: int


class EnergyAmplifierBuff(GenshinModel):
    """An energy amplifier challenge buff"""

    id: int
    name: str
    quality: int
    description: str = Field(galias="desc")
    energy: int


class EnergyAmplifierChallenge(GenshinModel):
    """An energy amplifier challenge"""

    id: int = Field(galias="challenge_id")
    name: str = Field(galias="challenge_name")
    characters: List[EnergyAmplifierCharacter] = Field(galias="avatars")
    energy: int
    difficulty: int
    high_score: int = Field(galias="max_score")
    criteria: List[EnergyAmplifierCriteria] = Field(galias="limit_conditions")
    multiplier: int = Field(galias="score_multiple")
    buffs: List[EnergyAmplifierBuff]


class EnergyAmplifier(GenshinModel):
    """An energy amplifier event"""

    start_time: datetime
    end_time: datetime
    challenges: List[EnergyAmplifierChallenge] = Field(alias="records")


# ---------------------------------------------------------
# A study in potions:


class PotionCharacter(BaseCharacter):
    """A study in potions character"""

    level: int
    trial: bool = Field(galias="is_trial")


class PotionPotion(GenshinModel):
    """A study in potions potion"""

    id: int
    name: str
    quality: int
    quality_icon: str = Field(galias="icon")
    mark_icon: str = Field(galias="cornor_mark")
    description: str = Field(galias="desc")


class PotionStage(GenshinModel):
    """A study in potions stage"""

    name: str = Field(galias="level_name")
    score: int
    difficulty: int
    difficulty_factor: int = Field(galias="factor")
    characters: List[PotionCharacter] = Field(galias="avatars")
    potions: List[PotionPotion] = Field(galias="buffs")


class PotionChallenge(GenshinModel):
    """A study in potions challenge"""

    name: str = Field(galias="stage_name")
    score: int = Field(galias="stage_score")
    finished: bool

    stages: List[PotionStage] = Field(galias="levels")


class Potions(GenshinModel):
    """A study in potions event"""

    challenges: List[PotionChallenge] = Field(galias="records")


# ---------------------------------------------------------
# Chinese activities:


class ChineseActivity(GenshinModel):
    """An arbitrary activty for chinese events"""

    start_time: datetime
    end_time: datetime
    total_score: int
    total_times: int
    records: List[Any]


# ---------------------------------------------------------
# Activities:


class Activities(GenshinModel):
    """A collection of genshin activities"""

    hyakunin_ikki: Optional[HyakuninIkki] = Field(None, gslug="sumo")
    labyrinth_warriors: Optional[LabyrinthWarriors] = Field(None, gslug="rogue")
    energy_amplifier: Optional[EnergyAmplifier] = Field(None, gslug="channeller_slab_copy")
    potions: Optional[Potions] = Field(None, gslug="potion")

    @root_validator(pre=True)
    def __flatten_activities(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if not values.get("activities"):
            return values

        slugs = {
            field.field_info.extra["gslug"]: name
            for name, field in cls.__fields__.items()
            if field.field_info.extra.get("gslug")
        }

        for activity in values["activities"]:
            for name, value in activity.items():
                if "exists_data" not in value:
                    continue

                name = slugs.get(name, name)
                values[name] = value if value["exists_data"] else None

        return values



import re
from enum import IntEnum
from typing import Any, Dict, List

from pydantic import Field, validator

from .base import GenshinModel

__all__ = [
    "GenshinAccount",
    "RecordCardData",
    "RecordCardSetting",
    "RecordCard",
    "Gender",
    "SearchUser",
]


class GenshinAccount(GenshinModel):
    """A genshin account"""

    uid: int = Field(galias="game_uid")
    level: int
    nickname: str
    server: str = Field(galias="region")
    server_name: str = Field(galias="region_name")


class RecordCardData(GenshinModel):
    """A data entry of a record card"""

    name: str
    value: str


class RecordCardSetting(GenshinModel):
    """A privacy setting of a record card"""

    id: int = Field(galias="switch_id")
    description: str = Field(galias="switch_name")
    public: bool = Field(galias="is_public")

    @property
    def name(self) -> str:
        return {
            1: "Battle Chronicle",
            2: "Character Details",
            3: "Real-Time Notes",
        }.get(self.id, "")


class RecordCard(GenshinAccount):
    """A genshin record card containing very basic user info"""

    uid: int = Field(galias="game_role_id")

    data: List[RecordCardData]
    privacy_settings: List[RecordCardSetting] = Field(galias="data_switches")

    # unknown meaning
    background_image: str
    has_uid: bool = Field(galias="has_role")
    public: bool = Field(galias="is_public")

    @property
    def days_active(self) -> int:
        return int(self.data[0].value)

    @property
    def characters(self) -> int:
        return int(self.data[1].value)

    @property
    def achievements(self) -> int:
        return int(self.data[2].value)

    @property
    def spiral_abyss(self) -> str:
        return self.data[3].value

    @property
    def html_url(self) -> str:
        """The html url"""
        return f"https://webstatic-sea.hoyolab.com/app/community-game-records-sea/index.html?uid={self.uid}#/ys"

    def as_dict(self) -> Dict[str, Any]:
        """Helper function which turns fields into properly named ones"""
        return {d.name: (int(d.value) if d.value.isdigit() else d.value) for d in self.data}


class Gender(IntEnum):
    unknown = 0
    male = 1
    female = 2
    other = 3


class SearchUser(GenshinModel):
    """A user from a search result"""

    hoyolab_uid: int = Field(galias="uid")
    nickname: str
    introduction: str = Field(galias="introduce")
    avatar_id: int = Field(galias="avatar")
    gender: Gender
    icon: str = Field(galias="avatar_url")

    @validator("nickname")
    def __remove_tag(cls, v: str) -> str:
        return re.sub(r"<.+?>", "", v)



from datetime import datetime
from typing import Any, Dict, List, Literal

from pydantic import Field, root_validator

from .base import BaseCharacter, GenshinModel

__all__ = [
    "AbyssRankCharacter",
    "AbyssCharacter",
    "CharacterRanks",
    "Battle",
    "Chamber",
    "Floor",
    "SpiralAbyss",
    "SpiralAbyssPair",
]


class AbyssRankCharacter(BaseCharacter):
    """A character with a value of a rank"""

    id: int = Field(galias="avatar_id")
    icon: str = Field(galias="avatar_icon")

    value: int


class AbyssCharacter(BaseCharacter):
    """A character with just a level"""

    level: int


class CharacterRanks(GenshinModel):
    """A collection of rankings achieved during spiral abyss runs"""

    # fmt: off
    most_played: List[AbyssRankCharacter] = Field([],      galias="reveal_rank",       mi18n="bbs/go_fight_count")
    most_kills: List[AbyssRankCharacter] = Field([],       galias="defeat_rank",       mi18n="bbs/max_rout_count")
    strongest_strike: List[AbyssRankCharacter] = Field([], galias="damage_rank",       mi18n="bbs/powerful_attack")
    most_damage_taken: List[AbyssRankCharacter] = Field([],galias="take_damage_rank",  mi18n="bbs/receive_max_damage")
    most_bursts_used: List[AbyssRankCharacter] = Field([], galias="energy_skill_rank", mi18n="bbs/element_break_count")
    most_skills_used: List[AbyssRankCharacter] = Field([], galias="normal_skill_rank", mi18n="bbs/element_skill_use_count")
    # fmt: on

    def as_dict(self, lang: str = "en-us") -> Dict[str, Any]:
        """Helper function which turns fields into properly named ones"""
        return {
            self._get_mi18n(field, lang): getattr(self, field.name)
            for field in self.__fields__.values()
        }


class Battle(GenshinModel):
    """A battle in the spiral abyss"""

    half: int = Field(galias="index")
    timestamp: datetime
    characters: List[AbyssCharacter] = Field(galias="avatars")


class Chamber(GenshinModel):
    """A chamber of the spiral abyss"""

    chamber: int = Field(galias="index")
    stars: int = Field(galias="star")
    max_stars: Literal[3] = Field(galias="max_star")
    battles: List[Battle]


class Floor(GenshinModel):
    """A floor of the spiral abyss"""

    floor: int = Field(galias="index")
    # icon: str - unused
    # settle_time: int - appsample might be using this?
    unlocked: Literal[True] = Field(galias="is_unlock")
    stars: int = Field(galias="star")
    max_stars: Literal[9] = Field(galias="max_star")  # maybe one day
    chambers: List[Chamber] = Field(galias="levels")


class SpiralAbyss(GenshinModel):
    """Information about Spiral Abyss runs during a specific season"""

    unlocked: bool = Field(galias="is_unlock")
    season: int = Field(galias="schedule_id")
    start_time: datetime
    end_time: datetime

    total_battles: int = Field(galias="total_battle_times")
    total_wins: str = Field(galias="total_win_times")
    max_floor: str
    total_stars: int = Field(galias="total_star")

    ranks: CharacterRanks

    floors: List[Floor]

    @root_validator(pre=True)
    def __nest_ranks(cls, values: Dict[str, Any]) -> Dict[str, AbyssCharacter]:
        """By default ranks are for some reason on the same level as the rest of the abyss"""
        values.setdefault("ranks", {}).update(values)
        return values


class SpiralAbyssPair(GenshinModel):
    """A pair of both current and previous spiral abyss

    This may not be a namedtuple due to how pydantic handles them
    """

    current: SpiralAbyss
    previous: SpiralAbyss



from typing import Dict, List

from pydantic import Field, validator

from .base import GenshinModel, PartialCharacter, Unique

__all__ = [
    "PartialCharacter",
    "Weapon",
    "ArtifactSetEffect",
    "ArtifactSet",
    "Artifact",
    "Constellation",
    "Outfit",
    "Character",
]


class Weapon(GenshinModel, Unique):
    """A character's equipped weapon"""

    id: int
    icon: str
    name: str
    rarity: int
    description: str = Field(galias="desc")
    level: int
    type: str = Field(galias="type_name")
    ascension: int = Field(galias="promote_level")
    refinement: int = Field(galias="affix_level")


class ArtifactSetEffect(GenshinModel):
    """An effect of an artifact set"""

    pieces: int = Field(galias="activation_number")
    effect: str
    enabled: bool = False

    class Config:
        # this is for the "enabled" field, hopefully nobody abuses this
        allow_mutation = True


class ArtifactSet(GenshinModel, Unique):
    """An artifact set"""

    id: int
    name: str
    effects: List[ArtifactSetEffect] = Field(galias="affixes")


class Artifact(GenshinModel, Unique):
    """A character's equipped artifact"""

    id: int
    icon: str
    name: str
    pos_name: str
    pos: int
    rarity: int
    level: int
    set: ArtifactSet


class Constellation(GenshinModel, Unique):
    """A character constellation"""

    id: int
    icon: str
    pos: int
    name: str
    effect: str
    activated: bool = Field(galias="is_actived")

    @property
    def scaling(self) -> bool:
        """Whether the constellation is simply for talent scaling"""
        return "U" in self.icon


class Outfit(GenshinModel, Unique):
    """An unlocked outfit of a character"""

    id: int
    icon: str
    name: str


class Character(PartialCharacter):
    """A character with equipment"""

    weapon: Weapon
    artifacts: List[Artifact] = Field(galias="reliquaries")
    constellations: List[Constellation]
    outfits: List[Outfit] = Field(galias="costumes")

    @validator("artifacts")
    def __add_artifact_effect_enabled(cls, artifacts: List[Artifact]) -> List[Artifact]:
        sets: Dict[int, List[Artifact]] = {}
        for arti in artifacts:
            sets.setdefault(arti.set.id, []).append(arti)

        for artifact in artifacts:
            for effect in artifact.set.effects:
                if effect.pieces <= len(sets[artifact.set.id]):
                    effect.enabled = True

        return artifacts



"""Paginators for abstracting paginated resources"""
from __future__ import annotations

import abc
import asyncio
import heapq
from datetime import datetime
from typing import *

from .models import *
from .utils import aislice, amerge

if TYPE_CHECKING:
    from .client import ChineseClient, GenshinClient

__all__ = [
    "BasePaginator",
    "DailyRewardPaginator",
    "ChineseDailyRewardPaginator",
    "DiaryPaginator",
    "IDPagintor",
    "AuthkeyPaginator",
    "WishHistory",
    "Transactions",
    "MergedPaginator",
    "MergedWishHistory",
    "MergedTransactions",
]


class _Model(Protocol):
    id: int
    time: datetime


T = TypeVar("T", bound=Any)
MT = TypeVar("MT", bound=_Model, covariant=True)
TransactionT = TypeVar("TransactionT", bound=BaseTransaction, covariant=True)


class BasePaginator(abc.ABC):
    """A base paginator requiring the implementation of some methods"""

    @property
    @abc.abstractmethod
    def exhausted(self) -> bool:
        ...

    @abc.abstractmethod
    async def next_page(self) -> Any:
        ...

    @abc.abstractmethod
    def __aiter__(self) -> AsyncIterator[Any]:
        ...

    async def flatten(self) -> Sequence[Any]:
        return [item async for item in self]

    def __await__(self) -> Generator[Any, Any, Sequence[Any]]:
        return self.flatten().__await__()


class DailyRewardPaginator(BasePaginator):
    """A paginator specifically for claimed daily rewards"""

    client: GenshinClient
    limit: Optional[int]
    lang: Optional[str]
    current_page: Optional[int]

    page_size: int = 10

    def __init__(self, client: GenshinClient, limit: int = None, lang: str = None) -> None:
        """Create a new daily reward pagintor

        :param client: A client for making http requests
        :param limit: The maximum amount of rewards to get
        :param lang: The language to use - currently broken
        """
        self.client = client
        self.limit = limit
        self.lang = lang

        self.current_page = 1

    @property
    def exhausted(self) -> bool:
        """Whether all pages have been fetched"""
        return self.current_page is None

    def __repr__(self) -> str:
        return f"{type(self).__name__}(limit={self.limit})"

    async def _get_page(self, page: int) -> List[ClaimedDailyReward]:
        params = dict(current_page=page)
        data = await self.client.request_daily_reward("award", params=params, lang=self.lang)
        return [ClaimedDailyReward(**i) for i in data["list"]]

    async def next_page(self) -> List[ClaimedDailyReward]:
        """Get the next page of the paginator"""
        if self.current_page is None:
            raise Exception("No more pages")

        data = await self._get_page(self.current_page)

        if len(data) < self.page_size:
            self.current_page = None
            return data

        self.current_page += 1
        return data

    async def _iter(self) -> AsyncIterator[ClaimedDailyReward]:
        """Iterate over pages until the end"""
        while not self.exhausted:
            page = await self.next_page()
            for i in page:
                yield i

    def __aiter__(self) -> AsyncIterator[ClaimedDailyReward]:
        """Iterate over all pages until the limit is reached"""
        return aislice(self._iter(), self.limit)


class ChineseDailyRewardPaginator(DailyRewardPaginator):
    """A paginator specifically for claimed daily rewards on chinese bbs"""

    client: ChineseClient
    uid: Optional[int]
    limit: Optional[int]
    current_page: Optional[int]

    page_size: int = 10

    def __init__(self, client: ChineseClient, uid: int = None, limit: int = None) -> None:
        """Create a new daily reward pagintor

        :param client: A client for making http requests
        :param uid: Genshin uid of the currently logged-in user
        :param limit: The maximum amount of rewards to get
        """
        self.client = client
        self.uid = uid
        self.limit = limit

    async def _get_page(self, page: int) -> List[ClaimedDailyReward]:
        params = dict(current_page=page)
        data = await self.client.request_daily_reward("award", self.uid, params=params)
        return [ClaimedDailyReward(**i) for i in data["list"]]


class DiaryPaginator(BasePaginator):
    """A paginator specifically for diary entries"""

    # TODO: Option to merge all months together
    client: GenshinClient

    uid: Optional[int]
    type: int
    month: Optional[int]
    limit: Optional[int]
    lang: Optional[str]

    _data: Optional[BaseDiary] = None
    page_size: int = 10

    def __init__(
        self,
        client: GenshinClient,
        uid: int = None,
        type: int = 1,
        month: int = None,
        limit: int = None,
        lang: str = None,
    ) -> None:
        """Create a new daily reward pagintor

        :param client: A client for making http requests
        :param uid: Genshin uid of the currently logged-in user
        :param type: The type of currency to get
        :param month: The month in the year to see the history for
        :param limit: The maximum amount of actions to get
        :param lang: The language to use
        """
        self.client = client
        self.uid = uid
        self.type = type
        self.month = month
        self.limit = limit
        self.lang = lang

        self.current_page = 1

    @property
    def exhausted(self) -> bool:
        """Whether all pages have been fetched"""
        return self.current_page is None

    def __repr__(self) -> str:
        return f"{type(self).__name__}(uid={self.uid}, type={self.type}, limit={self.limit})"

    async def _get_page(self, page: int) -> DiaryPage:
        data = await self.client.request_ledger(
            self.uid,
            detail=True,
            month=self.month,
            lang=self.lang,
            params=dict(type=self.type, current_page=page, limit=10),
        )
        return DiaryPage(**data)

    async def next_page(self) -> List[DiaryAction]:
        """Get the next page of the paginator"""
        if self.current_page is None:
            raise Exception("No more pages")

        page = await self._get_page(self.current_page)
        self._data = page

        actions = page.actions

        if len(actions) < self.page_size:
            self.current_page = None
            return actions

        self.current_page += 1
        return actions

    async def _iter(self) -> AsyncIterator[DiaryAction]:
        """Iterate over pages until the end"""
        while not self.exhausted:
            page = await self.next_page()
            for i in page:
                yield i

    def __aiter__(self) -> AsyncIterator[DiaryAction]:
        """Iterate over all pages until the limit is reached"""
        return aislice(self._iter(), self.limit)

    @property
    def data(self) -> BaseDiary:
        """Get data bound to the diary

        This requires at least one page to have been fetched
        """
        if self._data is None:
            raise RuntimeError("At least one item must be fetched before data can be gotten")

        return self._data


class IDPagintor(BasePaginator, Generic[MT]):
    """A paginator of genshin end_id pages"""

    __repr_args__: Sequence[str] = ["limit"]

    client: GenshinClient
    limit: Optional[int]
    end_id: Optional[int]

    page_size: int = 20

    def __init__(self, client: GenshinClient, *, limit: int = None, end_id: int = 0) -> None:
        """Create a new paginator

        :param client: A client for making http requests
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        self.client = client
        self.limit = limit
        self.end_id = end_id

    @property
    def exhausted(self) -> bool:
        """Whether all pages have been fetched"""
        return self.end_id is None

    def __repr__(self) -> str:
        args = ", ".join(f"{i}={getattr(self, i)!r}" for i in self.__repr_args__)
        return f"{type(self).__name__}({args})"

    async def _get_page(self, end_id: int) -> List[MT]:
        raise NotImplementedError

    def _cache_key(self, end_id: int) -> Tuple[int, str]:
        return (end_id, "")

    def _update_cache(self, data: List[MT]) -> bool:
        if self.client.paginator_cache is None:
            return False

        cache = self.client.paginator_cache

        if self.end_id:
            cache[self._cache_key(self.end_id)] = data[0]

        for p, n in zip(data, data[1:]):
            cache[self._cache_key(p.id)] = n

        return True

    def _collect_cache(self) -> Iterator[MT]:
        cache = self.client.paginator_cache
        if cache is None or self.end_id is None:
            return

        key = self._cache_key(self.end_id)
        while key in cache:
            yield cache[key]
            self.end_id = cache[key].id
            key = self._cache_key(self.end_id)

    async def next_page(self) -> List[MT]:
        """Get the next page of the paginator"""
        if self.end_id is None:
            raise Exception("No more pages")

        data = await self._get_page(self.end_id)

        self._update_cache(data)

        # mark paginator as exhausted
        if len(data) < self.page_size:
            self.end_id = None
            return data

        self.end_id = data[-1].id
        return data

    async def _iter(self) -> AsyncIterator[MT]:
        """Iterate over pages until the end"""
        # tfw no yield from in asyn iterators
        while self.end_id is not None:
            for i in self._collect_cache():
                yield i

            page = await self.next_page()
            for i in page:
                yield i

    def __aiter__(self) -> AsyncIterator[MT]:
        """Iterate over all pages until the limit is reached"""
        return aislice(self._iter(), self.limit)

    async def flatten(self, *, lazy: bool = True) -> List[MT]:
        """Flatten the entire iterator into a list

        :param lazy: Limit to only one request at a time
        """
        return [item async for item in self]

    async def first(self) -> MT:
        """Get the very first item"""
        x = await self._iter().__anext__()
        self.end_id = None  # invalidate the iterator
        return x


class AuthkeyPaginator(IDPagintor[MT]):
    """A paginator which utilizes authkeys"""

    __repr_args__ = ["limit", "lang"]

    _authkey: Optional[str]
    _lang: Optional[str]

    def __init__(
        self,
        client: GenshinClient,
        lang: str = None,
        authkey: str = None,
        limit: int = None,
        end_id: int = 0,
    ) -> None:
        """Create a new authkey paginator

        :param client: A client for making http requests
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        super().__init__(client, limit=limit, end_id=end_id)
        self._lang = lang
        self._authkey = authkey

    @property
    def lang(self) -> str:
        """The language used for fetching data"""
        return self._lang or self.client.lang

    @property
    def authkey(self) -> str:
        """The authkey used for fetching data"""
        authkey = self._authkey or self.client.authkey
        if authkey is None:
            raise RuntimeError("No authkey set for client")

        return authkey


class WishHistory(AuthkeyPaginator[Wish]):
    """A paginator for wishes"""

    __repr_args__ = ["banner_type", "limit", "lang"]

    client: GenshinClient
    banner_type: BannerType

    def __init__(self, client: GenshinClient, banner_type: BannerType, **kwargs: Any) -> None:
        """Create a new wish history paginator

        :param client: A client for making http requests
        :param banner_type: The banner from which to get the wishes
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        super().__init__(client, **kwargs)
        if banner_type not in [100, 200, 301, 302]:
            raise ValueError(f"Invalid banner type: {banner_type!r}")
        self.banner_type = banner_type

    def _cache_key(self, end_id: int) -> Tuple[int, str]:
        return (end_id, self.lang)

    async def _get_banner_name(self) -> str:
        """Get the banner name of banner_type"""
        banner_types = await self.client.get_banner_names(lang=self._lang, authkey=self._authkey)
        return banner_types[self.banner_type]

    async def _get_page(self, end_id: int) -> List[Wish]:
        data = await self.client.request_gacha_info(
            "getGachaLog",
            lang=self._lang,
            authkey=self._authkey,
            params=dict(gacha_type=self.banner_type, size=self.page_size, end_id=end_id),
        )
        banner_name = await self._get_banner_name()
        return [Wish(**i, banner_name=banner_name) for i in data["list"]]


class Transactions(AuthkeyPaginator[TransactionT]):
    """A paginator for transactions"""

    __repr_args__ = ["kind", "limit", "lang"]

    client: GenshinClient
    kind: TransactionKind

    def __init__(self, client: GenshinClient, kind: TransactionKind, **kwargs: Any) -> None:
        """Create a new transaction paginator

        :param client: A client for making http requests
        :param kind: The kind of transactions to get
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        super().__init__(client, **kwargs)
        if kind not in ["primogem", "crystal", "resin", "artifact", "weapon"]:
            raise ValueError(f"Invalid transaction kind: {kind}")
        self.kind = kind

    def _cache_key(self, end_id: int) -> Tuple[int, str]:
        return (end_id, self.lang)

    async def _get_page(self, end_id: int):
        endpoint = "get" + self.kind.capitalize() + "Log"

        data, reasons = await asyncio.gather(
            self.client.request_transaction(
                endpoint,
                lang=self._lang,
                authkey=self._authkey,
                params=dict(end_id=end_id, size=20),
            ),
            self.client._get_transaction_reasons(self.lang),
        )

        transactions = []
        for trans in data["list"]:
            cls = ItemTransaction if "name" in trans else Transaction
            reason_id = trans["reason"]
            trans = {**trans, "reason_id": reason_id, "reason": reasons.get(reason_id, "")}
            transactions.append(cls(**trans, kind=self.kind))

        return transactions


class MergedPaginator(AuthkeyPaginator[MT]):
    """A paginator merging a collection of other paginators"""

    _paginators: List[IDPagintor[MT]]

    def __init__(self, client: GenshinClient, **kwargs: Any) -> None:
        super().__init__(client, **kwargs)

    def next_page(self) -> NoReturn:
        raise RuntimeError('Merged paginators cannot be gotten by "pages"')

    def _key(self, item: _Model) -> float:
        return item.time.timestamp()

    def _iter(self) -> AsyncIterator[MT]:
        return amerge(self._paginators, key=self._key)

    async def flatten(self, *, lazy: bool = False) -> List[MT]:
        if self.limit is not None and lazy:
            it = aislice(amerge(self._paginators, key=self._key), self.limit)
            return [x async for x in it]

        coros = (p.flatten() for p in self._paginators)
        lists = await asyncio.gather(*coros)
        return list(heapq.merge(*lists, key=self._key))[: self.limit]


class MergedWishHistory(MergedPaginator[Wish]):
    """A paginator for wish histories from multiple banners"""

    __repr_args__ = ["banner_types", "limit", "lang"]

    client: GenshinClient
    banner_types: List[BannerType]

    def __init__(
        self,
        client: GenshinClient,
        banner_types: List[BannerType] = None,
        **kwargs: Any,
    ) -> None:
        """Create a new merged wish history paginator

        :param client: A client for making http requests
        :param banner_types: The banners from which to get the wishes
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        super().__init__(client, **kwargs)
        self.banner_types = banner_types or [100, 200, 301, 302]

        self._paginators = [WishHistory(client, b, **kwargs) for b in self.banner_types]

    async def flatten(self, *, lazy: bool = False) -> List[Wish]:
        # before we gather all histories we should get the banner name
        asyncio.create_task(self.client.get_banner_names(lang=self._lang, authkey=self._authkey))
        return await super().flatten(lazy=lazy)


class MergedTransactions(MergedPaginator[BaseTransaction]):
    """A paginator for transactions from multiple banners"""

    __repr_args__ = ["kinds", "limit", "lang"]

    client: GenshinClient
    kinds: List[TransactionKind]

    def __init__(
        self,
        client: GenshinClient,
        kinds: List[TransactionKind] = None,
        **kwargs: Any,
    ) -> None:
        """Create a new transaction paginator

        :param client: A client for making http requests
        :param kinds: The kinds of transactions to get
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param limit: The maximum amount of rewards to get
        :param end_id: The ending id to start getting data from
        """
        super().__init__(client, **kwargs)
        self.kinds = kinds or ["primogem", "crystal", "resin", "artifact", "weapon"]

        self._paginators = [Transactions(client, kind, **kwargs) for kind in self.kinds]



"""Utility functions for extracting data from logfiles"""
import os
import re
from typing import List, Optional
from urllib.parse import unquote

from .misc import get_tempdir

__all__ = ["get_logfile", "extract_authkey", "get_authkey", "get_banner_ids"]

AUTHKEY_FILE = os.path.join(get_tempdir(), "genshin_authkey.txt")


def get_logfile() -> Optional[str]:
    """Find a Genshin Impact logfile

    :returns: A logfile path or None if no genshin installation exists
    """
    mihoyo_dir = os.path.expanduser("~/AppData/LocalLow/miHoYo/")
    for name in ["Genshin Impact", "原神", "YuanShen"]:
        output_log = os.path.join(mihoyo_dir, name, "output_log.txt")
        if os.path.isfile(output_log):
            return output_log

    return None  # no genshin installation


def _read_logfile(logfile: Optional[str] = None) -> str:
    """Returns the contents of a logfile

    :param logfile: A path to a logfile
    """
    logfile = logfile or get_logfile()
    if logfile is None:
        raise FileNotFoundError("No Genshin Installation was found, could not get gacha data.")
    with open(logfile) as file:
        return file.read()


def extract_authkey(string: str) -> Optional[str]:
    """Extracts an authkey from the provided string

    :param string: A string containing a url with an authkey query parameter
    :returns: An extracted and parsed authkey if found
    """
    match = re.search(r"https://.+?authkey=([^&#]+)", string, re.MULTILINE)
    if match is not None:
        return unquote(match.group(1))
    return None


def get_authkey(logfile: str = None) -> str:
    """Get an authkey contained in a logfile

    :param logfile: A path to a logfile
    :returns: An extracted and parsed authkey if found
    :raises ValueError: Authkey could not be found
    """
    # first try the log
    authkey = extract_authkey(_read_logfile(logfile))
    if authkey is not None:
        with open(AUTHKEY_FILE, "w") as file:
            file.write(authkey)
        return authkey

    # otherwise try the tempfile (may be expired!)
    if os.path.isfile(AUTHKEY_FILE):
        with open(AUTHKEY_FILE) as file:
            return file.read()

    raise ValueError(
        "No authkey could be found in the logs or in a tempfile. "
        "Open the history in-game first before attempting to request it."
    )


def get_banner_ids(logfile: str = None) -> List[str]:
    """Gets all banner ids from a log file.
    You need to open the details of all banners for this to work.

    :param logfile: A path to a logfile
    """
    log = _read_logfile(logfile)
    ids = re.findall(r"OnGetWebViewPageFinish:https://.+?gacha_id=([^&#]+)", log)
    return list(set(ids))



from __future__ import annotations

import json
from datetime import datetime
from typing import Any, Dict, List, Tuple

from pydantic import Field, validator

from .base import GenshinModel

__all__ = ["MapInfoDetails", "MapInfo", "MapNode", "MapPoint", "MapLocation"]


class MapInfoDetails(GenshinModel):
    """Rendering details about a specific map"""

    slices: List[List[str]]
    origin: Tuple[int, int]
    total_size: Tuple[int, int]
    padding: Tuple[int, int]

    @validator("slices", pre=True)
    def __flatten_slices(cls, v: List[List[Dict[str, Any]]]) -> List[List[str]]:
        return [[j["url"] for j in i] for i in v]


class MapInfo(GenshinModel):
    """Information about a specific map"""

    id: int
    name: str
    icon: str

    details: MapInfoDetails = Field(galias="detail")

    @validator("details", pre=True)
    def __parse_detail(cls, v: str) -> Any:
        return json.loads(v)


class MapNode(GenshinModel):
    """A label or category node"""

    id: int
    name: str
    icon: str
    parent_id: int
    depth: int
    node_type: int
    jump_type: int
    jump_target_id: int
    display_priority: int
    children: List[MapNode]
    activity_page_label: int
    area_page_label: List[int]
    is_all_area: bool

    @validator("name")
    def __remove_style(cls, v: str) -> str:
        return v.replace("\xa0", " ")


class MapPoint(GenshinModel):
    """A point on the map"""

    id: int
    label_id: int
    x: float = Field(galias="x_pos")
    y: float = Field(galias="y_pos")
    display_state: int

    ctime: datetime
    author: str = Field(galias="author_name")

    @property
    def pos(self) -> Tuple[float, float]:
        return self.x, self.y


class MapLocation(GenshinModel):
    """A general location on the map"""

    id: int
    name: str
    parent_id: int
    map_id: int

    lx: float = Field(galias="l_x")
    ly: float = Field(galias="l_y")
    rx: float = Field(galias="r_x")
    ry: float = Field(galias="r_y")

    children: List[MapLocation]


MapNode.update_forward_refs()
MapLocation.update_forward_refs()



from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal

from pydantic import Field, root_validator, validator

from .base import BaseCharacter, GenshinModel

__all__ = ["Expedition", "Notes"]


class Expedition(GenshinModel):
    """A Real-Time note expedition"""

    character: BaseCharacter = Field(galias="avatar_side_icon")
    status: Literal["Ongoing", "Finished"]
    completed_at: datetime

    @property
    def finished(self) -> bool:
        """Whether the expedition has finished"""
        return self.remaining_time == 0

    @property
    def remaining_time(self) -> float:
        """The remaining time until expedition completion in seconds"""
        remaining = self.completed_at - datetime.now().astimezone()
        return max(remaining.total_seconds(), 0)

    @root_validator(pre=True)
    def __process_timedelta(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        time = timedelta(seconds=int(values["remained_time"]))
        values["completed_at"] = datetime.now().astimezone() + time

        return values

    @validator("character", pre=True)
    def __complete_character(cls, v: str) -> BaseCharacter:
        return BaseCharacter(icon=v)


class Notes(GenshinModel):
    """Real-Time notes"""

    current_resin: int
    max_resin: int
    resin_recovered_at: datetime

    current_realm_currency: int = Field(galias="current_home_coin")
    max_realm_currency: int = Field(galias="max_home_coin")
    realm_currency_recovered_at: datetime = Field(galias="home_coin_recovery_time")

    completed_commissions: int = Field(galias="finished_task_num")
    max_comissions: int = Field(galias="total_task_num")
    claimed_comission_reward: bool = Field(galias="is_extra_task_reward_received")

    remaining_resin_discounts: int = Field(galias="remain_resin_discount_num")
    max_resin_discounts: int = Field(galias="resin_discount_num_limit")

    expeditions: List[Expedition]
    max_expeditions: int = Field(galias="max_expedition_num")

    @property
    def until_resin_recovery(self) -> float:
        """The remaining time until resin recovery in seconds"""
        remaining = self.resin_recovered_at - datetime.now().astimezone()
        return max(remaining.total_seconds(), 0)

    @property
    def until_realm_currency_recovery(self) -> float:
        """The remaining time until resin recovery in seconds"""
        remaining = self.realm_currency_recovered_at - datetime.now().astimezone()
        return max(remaining.total_seconds(), 0)

    @root_validator(pre=True)
    def __process_timedelta(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        now = datetime.now().astimezone()

        values["resin_recovered_at"] = now + timedelta(seconds=int(values["resin_recovery_time"]))

        t = timedelta(seconds=int(values["realm_currency_recovered_at"]))
        values["realm_currency_recovered_at"] = now + t

        return values



"""A client interacting directly with the api"""
from __future__ import annotations

import asyncio
import base64
import json as json_
import logging
import os
import uuid
import warnings
from datetime import datetime
from http.cookies import SimpleCookie
from typing import *
from urllib.parse import unquote

import aiohttp
from yarl import URL

from . import errors
from .constants import CHARACTER_NAMES, LANGS
from .models import *
from .paginator import *
from .utils import *

__all__ = [
    "GenshinClient",
    "MultiCookieClient",
    "ChineseClient",
    "ChineseMultiCookieClient",
]


class GenshinClient:
    """A simple http client for genshin endpoints

    :var logger: A logger used for debugging
    :var cache: A cache for http requests
    :var paginator_cache: A high-frequency access cache for paginators
    """

    DS_SALT = "6cqshh5dhw73bzxn20oexa9k516chk7s"
    ACT_ID = "e202102251931481"

    WEBSTATIC_URL = "https://webstatic-sea.hoyoverse.com/"
    TAKUMI_URL = "https://api-os-takumi.mihoyo.com/"
    RECORD_URL = "https://bbs-api-os.hoyoverse.com/game_record/"
    INFO_LEDGER_URL = "https://hk4e-api-os.hoyoverse.com/event/ysledgeros/month_info"
    DETAIL_LEDGER_URL = "https://hk4e-api-os.hoyoverse.com/event/ysledgeros/month_detail"
    CALCULATOR_URL = "https://sg-public-api.hoyoverse.com/event/calculateos/"
    REWARD_URL = "https://hk4e-api-os.hoyoverse.com/event/sol/"
    GACHA_INFO_URL = "https://hk4e-api-os.hoyoverse.com/event/gacha_info/api/"
    YSULOG_URL = "https://hk4e-api-os.hoyoverse.com/ysulog/api/"
    MAP_URL = "https://api-os-takumi-static.hoyoverse.com/common/map_user/ys_obc/v1/map/"
    STATIC_MAP_URL = "https://api-os-takumi-static.hoyoverse.com/common/map_user/ys_obc/v1/map"

    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"  # noqa: E501

    _session: Optional[aiohttp.ClientSession] = None
    _uid: Optional[int] = None
    logger: logging.Logger = logging.getLogger(__name__)

    cache: Optional[MutableMapping[Tuple[Any, ...], Any]] = None
    paginator_cache: Optional[MutableMapping[Tuple[Any, ...], Any]] = None

    fetched_mi18n: bool = False

    def __init__(
        self,
        cookies: Mapping[str, str] = None,
        authkey: str = None,
        *,
        lang: str = "en-us",
        debug: bool = False,
    ) -> None:
        """Create a new GenshinClient instance

        :param cookies: The cookies used for authenticaation
        :param authkey: The authkey used for paginators
        :param lang: The default language
        :param debug: Whether debug logs should be shown in stdout
        """
        if cookies:
            self.cookies = cookies

        self.authkey = authkey
        self.lang = lang
        self.debug = debug

    def __repr__(self) -> str:
        return f"<{type(self).__name__} lang={self.lang!r} hoyolab_uid={self.hoyolab_uid} debug={self.debug}>"

    # PROPERTIES:

    @property
    def session(self) -> aiohttp.ClientSession:
        """The current client session, created when needed"""
        if self._session is None:
            self._session = aiohttp.ClientSession()

        return self._session

    @property
    def cookies(self) -> Mapping[str, str]:
        """The cookie jar belonging to the current session"""
        return {cookie.key: cookie.value for cookie in self.session.cookie_jar}

    @cookies.setter
    def cookies(self, cookies: Mapping[str, Any]) -> None:
        cks = {str(key): value for key, value in cookies.items()}
        self.session.cookie_jar.clear()
        self.session.cookie_jar.update_cookies(cks)

    @property
    def hoyolab_uid(self) -> Optional[int]:
        """The logged-in user's hoyolab uid"""
        for cookie in self.session.cookie_jar:
            if cookie.key in ("ltuid", "account_id"):
                return int(cookie.value)

        return None

    @property
    def uid(self) -> Optional[int]:
        return self._uid

    @uid.setter
    def uid(self, uid: Any) -> None:
        if not str(uid).isdigit() or len(str(uid)) != 9:
            raise TypeError(f"Invalid uid: {uid}")

        self._uid = int(uid)

    @property
    def lang(self) -> str:
        """The default language, defaults to "en-us" """
        return self._lang

    @lang.setter
    def lang(self, lang: str) -> None:
        if lang not in LANGS:
            raise ValueError(f"{lang} is not a valid language, must be one of: " + ", ".join(LANGS))

        self._lang = lang

    @property
    def authkey(self) -> Optional[str]:
        """The default authkey"""
        return self._authkey

    @authkey.setter
    def authkey(self, authkey: Optional[str]) -> None:
        if authkey is not None:
            authkey = unquote(authkey)

            try:
                base64.b64decode(authkey, validate=True)
            except Exception as e:
                raise ValueError("authkey is not a valid base64 encoded string") from e

        self._authkey = authkey

    @property
    def debug(self) -> bool:
        """Whether the debug logs are being shown in stdout"""
        return logging.getLogger("genshin").level == logging.DEBUG

    @debug.setter
    def debug(self, debug: bool) -> None:
        logging.basicConfig()
        level = logging.DEBUG if debug else logging.NOTSET
        logging.getLogger("genshin").setLevel(level)

    def set_cookies(
        self, cookies: Union[Mapping[str, Any], str] = None, **kwargs: Any
    ) -> Mapping[str, str]:
        """Helper cookie setter that accepts cookie headers

        :returns: The new cookies
        """
        if not bool(cookies) ^ bool(kwargs):
            raise TypeError("Cannot use both positional and keyword arguments at once")

        cookies = cookies or kwargs
        cookies = {morsel.key: morsel.value for morsel in SimpleCookie(cookies).values()}
        self.cookies = cookies
        return self.cookies

    def set_browser_cookies(self, browser: str = None) -> Mapping[str, str]:
        """Extract cookies from your browser and set them as client cookies

        Avalible browsers: chrome, chromium, opera, edge, firefox

        :param browser: The browser to extract the cookies from
        :returns: The extracted cookies
        """
        self.cookies = get_browser_cookies(browser)
        return self.cookies

    def set_authkey(self, authkey: str = None) -> None:
        """Sets an authkey for wish & transaction logs

        :param authkey: An authkey, a url containing an authkey or a path towards a logfile
        :returns: The new authkey
        """
        if authkey is None or os.path.isfile(authkey):
            authkey = get_authkey(authkey)
        else:
            authkey = extract_authkey(authkey) or authkey

        self.authkey = authkey

    def set_cache(
        self,
        maxsize: int,
        strategy: Literal["FIFO", "LFU", "LRU", "MRU", "RR"] = "LRU",
        *,
        ttl: int = None,
        getsizeof: Callable[[Any], float] = None,
    ) -> MutableMapping[Any, Any]:
        """Create and set a new cache for http requests

        :param maxsize: The maximum size of the cache
        :param strategy: The cache strategy to use, defaults to Least-Recently-Used
        :param ttl: The time to live of items, only works with LRU caches
        :param getsizeof: Function that gets the size of any objecct, by default everything has size of 1
        :returns: The newly created cache
        """
        import cachetools

        if ttl:
            if strategy != "LRU":
                raise ValueError("TTL caches must use LRU")

            self.cache = cachetools.TTLCache(maxsize, ttl, getsizeof=getsizeof)
            return self.cache
        elif strategy == "TTL":
            raise ValueError("TTL caches should be set using the ttl kwarg")

        cls_name = strategy + "Cache"
        if not hasattr(cachetools, cls_name):
            raise ValueError(f"Invalid strategy: {strategy}")

        self.cache = getattr(cachetools, cls_name)(maxsize, getsizeof=getsizeof)
        return self.cache

    async def _check_cache(
        self,
        key: Tuple[Any, ...],
        check: Callable[[Any], bool] = None,
        *,
        lang: str = None,
    ) -> Optional[Any]:
        """Check the cache for any entries"""
        if self.cache is None:
            return None

        key = key + (lang or self.lang,)

        if key not in self.cache:
            return None

        data = self.cache[key]
        if check is None or check(data):
            return data

        del self.cache[key]

        return None

    async def _update_cache(
        self,
        data: Any,
        key: Tuple[Any, ...],
        check: Callable[[Any], bool] = None,
        *,
        lang: str = None,
    ) -> None:
        """Update the cache with a new entry"""
        if self.cache is None:
            return

        key = key + (lang or self.lang,)

        if check is not None and not check(data):
            return

        self.cache[key] = data

    async def _request_hook(
        self,
        method: str,
        url: Union[str, URL],
        *,
        params: Dict[str, Any] = None,
        json: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> None:
        """A hook before request, by default performs a debug log"""
        url = URL(url)
        if params:
            params = {k: v for k, v in params.items() if k != "authkey"}
            url = url.with_query(params)

        string = f"{method} {url}"
        if json:
            string += "\n" + json_.dumps(json, separators=(",", ":"))

        self.logger.debug(string)

    # ASYNCIO HANDLERS:

    async def close(self) -> None:
        """Close the underlying aiohttp session"""
        if not self.session.closed:
            await self.session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc_info):
        await self.close()

    # RAW HTTP REQUESTS:

    @handle_ratelimits()
    async def request(
        self,
        url: Union[str, URL],
        method: str = "GET",
        *,
        headers: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request and return a parsed json response"""
        asyncio.create_task(self._fetch_mi18n())

        headers = headers or {}
        headers["User-Agent"] = self.USER_AGENT

        await self._request_hook(method, url, headers=headers, **kwargs)

        async with self.session.request(method, url, headers=headers, **kwargs) as r:
            r.raise_for_status()
            data = await r.json()

        if data["retcode"] == 0:
            return data["data"]

        errors.raise_for_retcode(data)

    async def request_webstatic(
        self,
        url: Union[str, URL],
        *,
        headers: Dict[str, Any] = None,
        cache: bool = True,
        **kwargs: Any,
    ) -> Any:
        """Request a static json file"""
        url = URL(self.WEBSTATIC_URL).join(URL(url))

        data = get_from_static_cache(str(url))
        if data is not None:
            return data

        headers = headers or {}
        headers["user-agent"] = self.USER_AGENT

        async with self.session.get(url, headers=headers, **kwargs) as r:
            r.raise_for_status()
            data = await r.json()

        if cache:
            save_to_static_cache(str(url), data)

        return data

    async def request_hoyolab(
        self,
        endpoint: Union[str, URL],
        *,
        method: str = "GET",
        lang: str = None,
        cache: Tuple[Any, ...] = None,
        cache_check: Callable[[Any], bool] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards misc hoyolabs api

        Community related data
        """
        if cache:
            data = await self._check_cache(cache, cache_check, lang=lang)
            if data:
                return data

        if not self.cookies:
            raise RuntimeError("No cookies provided")
        if lang not in LANGS and lang is not None:
            raise ValueError(f"{lang} is not a valid language, must be one of: " + ", ".join(LANGS))

        url = URL(self.TAKUMI_URL).join(URL(endpoint))

        headers = {
            "x-rpc-app_version": "1.5.0",
            "x-rpc-client_type": "4",
            "x-rpc-language": lang or self.lang,
            "ds": generate_dynamic_secret(self.DS_SALT),
        }

        data = await self.request(url, method, headers=headers, **kwargs)

        if cache:
            await self._update_cache(data, cache, cache_check, lang=lang)

        return data

    async def request_game_record(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        cache: Tuple[Any, ...] = None,
        cache_check: Callable[[Any], bool] = None,
        **kwargs: Any,
    ):
        """Make a request towards the game record endpoint

        User stats related data
        """
        # this is simply just an alias for shorter request endpoints
        url = URL(self.RECORD_URL).join(URL(endpoint))

        return await self.request_hoyolab(
            url, method=method, cache=cache, cache_check=cache_check, **kwargs
        )

    async def request_ledger(
        self,
        uid: int = None,
        detail: bool = False,
        *,
        month: int = None,
        lang: str = None,
        params: Dict[str, Any] = None,
    ) -> Dict[str, Any]:
        """Make a request towards the ys ledger endpoint

        Traveler's diary related data
        """
        params = params or {}

        url = URL(self.DETAIL_LEDGER_URL if detail else self.INFO_LEDGER_URL)

        params.update(await self._complete_uid(uid))
        params["month"] = month or datetime.now().month
        params["lang"] = lang or self.lang

        return await self.request(url, params=params)

    async def request_calculator(
        self,
        endpoint: str,
        *,
        method: str = "POST",
        lang: str = None,
        params: Dict[str, Any] = None,
        json: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards the calculator endpoint

        Calculator database and resource calculation
        """
        params = params or {}
        json = json or {}

        if not self.cookies:
            raise RuntimeError("No cookies provided")

        url = URL(self.CALCULATOR_URL).join(URL(endpoint))

        if method == "GET":
            params["lang"] = lang or self.lang
            json = None
        else:
            json["lang"] = lang or self.lang

        return await self.request(url, method, params=params, json=json, **kwargs)

    async def request_daily_reward(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: str = None,
        params: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards the daily reward endpoint

        Daily reward claiming and history
        """
        params = params or {}

        if not self.cookies:
            raise RuntimeError("No cookies provided")

        url = URL(self.REWARD_URL).join(URL(endpoint))

        params["lang"] = lang or self.lang
        params["act_id"] = self.ACT_ID

        return await self.request(url, method, params=params, **kwargs)

    async def request_gacha_info(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: str = None,
        authkey: Optional[str] = None,
        params: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards the game info endpoint

        Wish history related data
        """
        params = params or {}
        authkey = authkey or self.authkey

        if authkey is None:
            raise RuntimeError("No authkey provided")

        base_url = URL(self.GACHA_INFO_URL)
        url = base_url.join(URL(endpoint))

        params["authkey_ver"] = 1
        params["authkey"] = unquote(authkey)
        params["lang"] = create_short_lang_code(lang or self.lang)

        return await self.request(url, method, params=params, **kwargs)

    async def request_transaction(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: str = None,
        authkey: Optional[str] = None,
        params: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards the transaction log endpoint

        Transaction related data
        """
        params = params or {}
        authkey = authkey or self.authkey

        if authkey is None:
            raise RuntimeError("No authkey provided")

        base_url = URL(self.YSULOG_URL)
        url = base_url.join(URL(endpoint))

        params["authkey_ver"] = 1
        params["sign_type"] = 2
        params["authkey"] = unquote(authkey)
        params["lang"] = create_short_lang_code(lang or self.lang)

        return await self.request(url, method, params=params, **kwargs)

    async def request_map(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: str = None,
        map_id: int = 2,
        static: bool = False,
        params: Dict[str, Any] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Make a request towards the map endpoint

        Interactive map related data
        """
        params = params or {}

        base_url = self.STATIC_MAP_URL if static else self.MAP_URL
        url = URL(base_url).join(URL(endpoint))

        params["map_id"] = map_id
        params["app_sn"] = "ys_obc"
        params["lang"] = lang or self.lang

        return await self.request(url, method, params=params, **kwargs)

    async def login_with_ticket(self, login_ticket: str = "") -> None:
        """Complete cookies using a login ticket"""
        url = "https://webapi-os.account.hoyoverse.com/Api/cookie_accountinfo_by_loginticket"
        async with self.session.get(url, params=dict(login_ticket=login_ticket)) as r:
            r.raise_for_status()

    # HOYOLAB:

    async def genshin_accounts(self, *, lang: str = None) -> List[GenshinAccount]:
        """Get the genshin accounts of the currently logged-in user

        :params lang: The language to use
        """
        # TODO: Account for honkai accounts

        # fmt: off
        data = await self.request_hoyolab(
            "binding/api/getUserGameRolesByCookie",
            lang=lang,
            cache=("accounts", self.hoyolab_uid)
        )
        # fmt: on
        return [GenshinAccount(**i) for i in data["list"]]

    async def search_users(self, keyword: str, *, lang: str = None) -> List[SearchUser]:
        """Search hoyolab users

        :param keyword: The keyword to search with
        :params lang: The language to use
        """
        data = await self.request_hoyolab(
            "community/search/wapi/search/user",
            lang=lang,
            params=dict(keyword=keyword, page_size=20),
            cache=("search", keyword),
        )
        return [SearchUser(**i["user"]) for i in data["list"]]

    async def set_visibility(self, public: bool) -> None:
        """Sets your data to public or private.

        :param public: Whether the data should now be public
        """
        await self.request_game_record(
            "genshin/wapi/publishGameRecord",
            method="POST",
            json=dict(is_public=public, game_id=2),
        )

    async def get_recommended_users(self, *, limit: int = 200) -> List[SearchUser]:
        """Get a list of recommended active users

        :param limit: The maximum amount of users to return
        """
        data = await self.request_hoyolab(
            "community/user/wapi/recommendActive",
            params=dict(page_size=limit),
        )
        return [SearchUser(**i["user"]) for i in data["list"]]

    async def redeem_code(self, code: str, uid: int = None, *, lang: str = None) -> None:
        """Redeems a gift code for the current user

        :param code: The code to redeem
        :param uid: The specific uid to redeem for
        :param lang: The language to use
        """
        # do note that this endpoint is very quirky, can't really make this pretty
        if uid is not None:
            server = recognize_server(uid)
            lang = create_short_lang_code(lang or self.lang)
            await self.request(
                "https://hk4e-api-os.hoyoverse.com/common/apicdkey/api/webExchangeCdkey",
                params=dict(
                    uid=uid,
                    region=server,
                    cdkey=code,
                    game_biz="hk4e_global",
                    lang=lang,
                ),
            )
            return

        accounts = [a for a in await self.genshin_accounts() if a.level >= 10]

        for i, account in enumerate(accounts):
            # there's a ratelimit of 1 request every 5 seconds
            if i:
                await asyncio.sleep(5)

            await self.redeem_code(code, account.uid, lang=lang)

    # GAME RECORD:

    async def _fetch_raw_user(self, uid: int, lang: str = None) -> Dict[str, Any]:
        """Low-level http method for fetching the game record index"""
        server = recognize_server(uid)
        data = await self.request_game_record(
            "genshin/api/index",
            lang=lang,
            params=dict(server=server, role_id=uid),
            cache=("user", uid),
        )
        return data

    async def _fetch_raw_characters(self, uid: int, *, lang: str = None) -> List[Dict[str, Any]]:
        """Low-level http method for fetching the game record characters

        Caching with characters is optimized
        """
        server = recognize_server(uid)
        data = await self.request_game_record(
            "genshin/api/character",
            method="POST",
            lang=lang,
            json=dict(role_id=uid, server=server),
            cache=("characters", uid),
        )

        return data["avatars"]

    async def get_record_card(self, hoyolab_uid: int = None, *, lang: str = None) -> RecordCard:
        """Get a user's record card

        :param hoyolab_uid: A hoyolab uid
        :param lang: The language to use
        """
        data = await self.request_game_record(
            "card/wapi/getGameRecordCard",
            lang=lang,
            params=dict(uid=hoyolab_uid or self.hoyolab_uid),
            cache=("record", hoyolab_uid),
            cache_check=lambda data: len(data["list"]) > 0,
        )
        cards = data["list"]
        if not cards:
            raise errors.DataNotPublic({"retcode": 10102})

        return RecordCard(**cards[0])

    async def get_user(self, uid: int, *, lang: str = None) -> UserStats:
        """Get a user's stats and characters

        :param uid: A Genshin uid
        :param character_ids: The ids of characters you want to fetch
        :param all_characters: Whether to get every single character a user has. Discouraged.
        :param lang: The language to use
        """
        data, characters = await asyncio.gather(
            self._fetch_raw_user(uid, lang=lang),
            self._fetch_raw_characters(uid, lang=lang),
        )
        data["avatars"] = characters

        return UserStats(**data)

    async def get_partial_user(self, uid: int, *, lang: str = None) -> PartialUserStats:
        """Helper function to get a user without any equipment

        :param uid: A Genshin uid
        :param lang: The language to use
        """
        data = await self._fetch_raw_user(uid, lang=lang)
        return PartialUserStats(**data)

    async def get_characters(self, uid: int, *, lang: str = None) -> List[Character]:
        """Helper function to fetch characters from just their ids

        :param uid: A Genshin uid
        :param lang: The language to use
        """
        data = await self._fetch_raw_characters(uid, lang=lang)
        return [Character(**i) for i in data]

    async def get_spiral_abyss(
        self, uid: int, *, previous: bool = False, lang: str = None
    ) -> SpiralAbyss:
        """Get spiral abyss runs

        :param uid: A Genshin uid
        :param previous: Whether to get the record of the previous spiral abyss
        :param lang: The language to use
        """
        server = recognize_server(uid)
        schedule_type = 2 if previous else 1
        data = await self.request_game_record(
            "genshin/api/spiralAbyss",
            lang=lang,
            params=dict(role_id=uid, server=server, schedule_type=schedule_type),
            cache=("abyss", uid, schedule_type),
        )
        return SpiralAbyss(**data)

    async def get_notes(self, uid: int, *, lang: str = None) -> Notes:
        """Get the real-time notes.

        :param uid: A Genshin uid
        :param lang: The language to use
        """
        server = recognize_server(uid)
        data = await self.request_game_record(
            "genshin/api/dailyNote",
            lang=lang,
            params=dict(server=server, role_id=uid),
        )
        return Notes(**data)

    async def get_activities(self, uid: int, *, lang: str = None) -> Activities:
        """Get activities

        :param uid: A Genshin uid
        :param lang: The language to use
        """
        server = recognize_server(uid)
        data = await self.request_game_record(
            "genshin/api/activities",
            lang=lang,
            params=dict(server=server, role_id=uid),
            cache=("activities", uid),
        )
        return Activities(**data)

    async def get_full_user(self, uid: int, *, lang: str = None) -> FullUserStats:
        """Get a user with all their possible data

        :param uid: A Genshin uid
        :param lang: The language to use
        """
        user, abyss1, abyss2, activities = await asyncio.gather(
            self.get_user(uid, lang=lang),
            self.get_spiral_abyss(uid, lang=lang, previous=False),
            self.get_spiral_abyss(uid, lang=lang, previous=True),
            self.get_activities(uid, lang=lang),
        )
        abyss = {"current": abyss1, "previous": abyss2}
        return FullUserStats(**user.dict(), abyss=abyss, activities=activities)

    async def set_top_characters(self, character_ids: List[int], *, uid: int = None) -> None:
        """Set the top 8 visible characters for the current user

        :param character_ids: IDs of characters to be shown
        :param uid: Genshin uid of the currently logged-in user
        """
        json = dict(avatar_ids=character_ids)
        json.update(await self._complete_uid(uid, uid_key="role_id", server_key="server"))
        await self.request_game_record(
            "genshin/api/character/top",
            method="POST",
            json=json,
        )

    # LEDGER:

    async def get_diary(self, uid: int = None, *, month: int = None, lang: str = None) -> Diary:
        """Get a traveler's diary with earning details for the month

        :param uid: Genshin uid of the currently logged-in user
        :param month: The month in the year to see the history for
        :param lang: The language to use
        """
        data = await self.request_ledger(uid, month=month, lang=lang)
        return Diary(**data)

    def diary_log(
        self,
        uid: int = None,
        *,
        mora: bool = False,
        month: int = None,
        limit: int = None,
        lang: str = None,
    ) -> DiaryPaginator:
        """Create a new daily reward pagintor

        :param client: A client for making http requests
        :param uid: Genshin uid of the currently logged-in user
        :param mora: Whether the type of currency should be mora instead of primogems
        :param month: The month in the year to see the history for
        :param limit: The maximum amount of actions to get
        :param lang: The language to use
        """
        type = 2 if mora else 1
        return DiaryPaginator(self, uid, type, month, limit, lang)

    # CALCULATOR

    def calculator(self, *, lang: str = None) -> CalculatorBuilder:
        """Calculate the resources needed to level up various equipment.

        :param lang: The language to use
        """
        return CalculatorBuilder(self, lang=lang)

    async def _execute_calculator(self, builder: CalculatorBuilder) -> CalculatorResult:
        """Create a request to calculate the results of a builder"""
        json = await builder.build()
        data = await self.request_calculator("compute", lang=builder.lang, json=json)
        return CalculatorResult(**data)

    async def _get_calculator_items(
        self,
        slug: str,
        filters: Dict[str, Any],
        query: str = None,
        *,
        is_all: bool = False,
        sync: Union[int, bool] = False,
        lang: str = None,
    ) -> List[Dict[str, Any]]:
        """Get all items of a specific slug from a calculator"""
        if query and any(isinstance(v, list) and v for v in filters.values()):
            raise TypeError("Cannot specify a query and filter at the same time")

        endpoint = f"sync/{slug}/list" if sync else f"{slug}/list"
        json: Dict[str, Any] = dict(page=1, size=69420, is_all=is_all, **filters)
        if query:
            json.update(keywords=query)
        if sync:
            json.update(await self._complete_uid(sync if sync > 1 else None))

        data = await self.request_calculator(
            endpoint,
            lang=lang,
            json=json,
        )
        return data["list"]

    async def get_calculator_characters(
        self,
        *,
        query: str = None,
        elements: Sequence[int] = None,
        weapon_types: Sequence[int] = None,
        include_traveler: bool = False,
        sync: Union[int, bool] = False,
        lang: str = None,
    ) -> List[CalculatorCharacter]:
        """Get all characters provided by the Enhancement Progression Calculator

        :param query: A query to use when searching; incompatible with other filters
        :param elements: The elements of returned characters - refer to `.models.CALCULATOR_ELEMENTS`
        :param weapon_types: The weapon types of returned characters - refer to `.models.CALCULATOR_WEAPON_TYPES`
        :param lang: The language to use
        """
        data = await self._get_calculator_items(
            "avatar",
            lang=lang,
            is_all=include_traveler,
            sync=sync,
            query=query,
            filters=dict(
                element_attr_ids=elements or [],
                weapon_cat_ids=weapon_types or [],
            ),
        )
        return [CalculatorCharacter(**i) for i in data]

    async def get_calculator_weapons(
        self,
        *,
        query: str = None,
        types: Sequence[int] = None,
        rarities: Sequence[int] = None,
        lang: str = None,
    ) -> List[CalculatorWeapon]:
        """Get all weapons provided by the Enhancement Progression Calculator

        :param query: A query to use when searching; incompatible with other filters
        :param types: The types of returned weapons - refer to `.models.CALCULATOR_WEAPON_TYPES`
        :param rarities: The rarities of returned weapons
        :param lang: The language to use
        """
        data = await self._get_calculator_items(
            "weapon",
            lang=lang,
            query=query,
            filters=dict(
                weapon_cat_ids=types or [],
                weapon_levels=rarities or [],
            ),
        )
        return [CalculatorWeapon(**i) for i in data]

    async def get_calculator_artifacts(
        self,
        *,
        query: str = None,
        pos: int = 1,
        rarities: Sequence[int] = None,
        lang: str = None,
    ) -> List[CalculatorArtifact]:
        """Get all artifacts provided by the Enhancement Progression Calculator

        :param query: A query to use when searching; incompatible with other filters
        :param pos: The slot position of the returned weapon
        :param rarities: The rarities of returned artifacts
        :param lang: The language to use
        """
        data = await self._get_calculator_items(
            "reliquary",
            lang=lang,
            query=query,
            filters=dict(
                reliquary_cat_id=pos,
                reliquary_levels=rarities or [],
            ),
        )
        return [CalculatorArtifact(**i) for i in data]

    async def get_character_details(
        self,
        character_id: int,
        *,
        uid: int = None,
        lang: str = None,
    ) -> CalculatorCharacterDetails:
        """Get the weapon, artifacts and talents of a character

        Not related to the Battle Chronicle. This data is always private.

        :param lang: The language to use
        """
        params = dict(avatar_id=character_id)
        params.update(await self._complete_uid(uid))

        data = await self.request_calculator(
            "sync/avatar/detail",
            method="GET",
            lang=lang,
            params=params,
        )
        return CalculatorCharacterDetails(**data)

    async def get_character_talents(
        self,
        character_id: int,
        *,
        lang: str = None,
    ) -> List[CalculatorTalent]:
        """Get the talents of a character

        This only gets the talent names, not their levels.
        Use `get_character_details` for precise information.

        :param lang: The language to use
        """
        data = await self.request_calculator(
            "avatar/skill_list",
            method="GET",
            lang=lang,
            params=dict(avatar_id=character_id),
        )
        return [CalculatorTalent(**i) for i in data["list"]]

    async def get_complete_artifact_set(
        self,
        artifact_id: int,
        *,
        lang: str = None,
    ) -> List[CalculatorArtifact]:
        """Get all other artifacts that share a set with any given artifact

        Doesn't return the artifact passed into this function.

        :param lang: The language to use
        """
        data = await self.request_calculator(
            "reliquary/set",
            method="GET",
            lang=lang,
            params=dict(reliquary_id=artifact_id),
        )
        return [CalculatorArtifact(**i) for i in data["reliquary_list"]]

    # DAILY REWARDS:

    async def get_reward_info(self, *, lang: str = None) -> DailyRewardInfo:
        """Get the daily reward info for the current user

        :param lang: The language to use
        """
        data = await self.request_daily_reward("info", lang=lang)
        return DailyRewardInfo(data["is_sign"], data["total_sign_day"])

    async def get_monthly_rewards(self, *, lang: str = None) -> List[DailyReward]:
        """Get a list of all availible rewards for the current month

        :param lang: The language to use
        """
        func = perm_cache(
            ("rewards", datetime.utcnow().month, lang or self.lang),
            self.request_daily_reward,
        )
        data = await func("home", lang=lang)
        return [DailyReward(**i) for i in data["awards"]]

    def claimed_rewards(self, *, limit: int = None, lang: str = None) -> DailyRewardPaginator:
        """Get all claimed rewards for the current user

        NOTE: Languages are currently broken,
        the language is based off the language used to claim the reward

        :param limit: The maximum amount of rewards to get
        :param lang: The language to use - currently broken
        """
        return DailyRewardPaginator(self, limit=limit, lang=lang)

    @overload
    async def claim_daily_reward(
        self, *, lang: str = None, reward: Literal[True] = ...
    ) -> DailyReward:
        ...

    @overload
    async def claim_daily_reward(self, *, lang: str = None, reward: Literal[False]) -> None:
        ...

    async def claim_daily_reward(
        self, *, lang: str = None, reward: bool = True
    ) -> Optional[DailyReward]:
        """Signs into hoyolab and claims the daily reward.

        :param lang: The language to use
        :param reward: Whether to also fetch the claimed reward
        """
        await self.request_daily_reward("sign", method="POST", lang=lang)

        if not reward:
            return None

        info, rewards = await asyncio.gather(
            self.get_reward_info(lang=lang),
            self.get_monthly_rewards(lang=lang),
        )
        return rewards[info.claimed_rewards - 1]

    # WISH HISTORY:

    @overload
    def wish_history(
        self,
        banner_type: Optional[List[BannerType]] = ...,
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> MergedWishHistory:
        ...

    @overload
    def wish_history(
        self,
        banner_type: BannerType,
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> WishHistory:
        ...

    def wish_history(
        self,
        banner_type: Union[BannerType, List[BannerType]] = None,
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> Union[WishHistory, MergedWishHistory]:
        """Get the wish history of a user

        :param banner_type: The banner(s) from which to get the wishes
        :param limit: The maximum amount of wishes to get
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param end_id: The ending id to start getting data from
        """
        cls = WishHistory if isinstance(banner_type, int) else MergedWishHistory
        return cls(
            self,
            banner_type,  # type: ignore
            lang=lang,
            authkey=authkey,
            limit=limit,
            end_id=end_id,
        )

    async def get_banner_names(
        self, *, lang: str = None, authkey: str = None
    ) -> Dict[BannerType, str]:
        """Get a list of banner names

        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        """
        func = perm_cache(("banners", lang or self.lang), self.request_gacha_info)
        data = await func(
            "getConfigList",
            lang=lang,
            authkey=authkey,
        )
        return {int(i["key"]): i["name"] for i in data["gacha_type_list"]}  # type: ignore

    async def _get_banner_details(self, banner_id: str, *, lang: str = None) -> BannerDetails:
        """Get details of a specific banner using its id

        :param banner_id: A banner id
        :param lang: The language to use
        """
        data = await self.request_webstatic(
            f"/hk4e/gacha_info/os_asia/{banner_id}/{lang or self.lang}.json"
        )
        return BannerDetails(**data, banner_id=banner_id)

    async def get_banner_details(
        self, banner_ids: List[str] = None, *, lang: str = None
    ) -> List[BannerDetails]:
        """Get all banner details at once in a batch

        :param banner_ids: A list of banner ids, implicitly fetched when not provided
        :param lang: The language to use
        """
        try:
            banner_ids = banner_ids or get_banner_ids()
        except FileNotFoundError:
            banner_ids = []

        if len(banner_ids) < 3:
            banner_ids = await self.fetch_banner_ids()

        data = await asyncio.gather(*(self._get_banner_details(i, lang=lang) for i in banner_ids))
        return list(data)

    async def get_gacha_items(
        self, *, server: str = "os_asia", lang: str = None
    ) -> List[GachaItem]:
        """Get the list of characters and weapons that can be gotten from the gacha.

        :param server: The server to request the items from
        :param lang: The language to use
        """
        data = await self.request_webstatic(
            f"/hk4e/gacha_info/{server}/items/{lang or self.lang}.json", cache=False
        )
        return [GachaItem(**i) for i in data]

    # TRANSACTIONS:

    async def _get_transaction_reasons(self, lang: str) -> Dict[str, str]:
        """Get a mapping of transaction reasons

        :param lang: The language to use
        """
        base = "https://mi18n-os.hoyoverse.com/webstatic/admin/mi18n/hk4e_global/"
        data = await self.request_webstatic(base + f"m02251421001311/m02251421001311-{lang}.json")

        return {
            k.split("_")[-1]: v
            for k, v in data.items()
            if k.startswith("selfinquiry_general_reason_")
        }

    @overload
    def transaction_log(
        self,
        kind: Optional[List[TransactionKind]] = ...,
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> MergedTransactions:
        ...

    @overload
    def transaction_log(
        self,
        kind: Literal["primogem", "crystal", "resin"],
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> Transactions[Transaction]:
        ...

    @overload
    def transaction_log(
        self,
        kind: Literal["artifact", "weapon"],
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> Transactions[ItemTransaction]:
        ...

    def transaction_log(
        self,
        kind: Union[TransactionKind, List[TransactionKind]] = None,
        *,
        limit: int = None,
        lang: str = None,
        authkey: str = None,
        end_id: int = 0,
    ) -> Union[Transactions[Any], MergedTransactions]:
        """Get the transaction log of a user

        :param kind: The kind(s) of transactions to get
        :param limit: The maximum amount of wishes to get
        :param lang: The language to use
        :param authkey: The authkey to use when requesting data
        :param end_id: The ending id to start getting data from
        """
        cls = Transactions if isinstance(kind, str) else MergedTransactions
        return cls(
            self,
            kind,  # type: ignore
            lang=lang,
            authkey=authkey,
            limit=limit,
            end_id=end_id,
        )

    # INTERACTIVE MAP:

    async def _get_map_pin_icons(self, map_id: int = 2, *, lang: str = None) -> Dict[int, str]:
        """Get the icons of pins"""
        data = await self.request_map("spot_kind/get_icon_list", lang=lang, map_id=map_id)
        return {i["id"]: i["url"] for i in data["icons"]}

    async def get_map_info(self, map_id: int = 2, *, lang: str = None) -> MapInfo:
        """Get info about an interactive map"""
        data = await self.request_map(
            "map/info",
            lang=lang,
            map_id=map_id,
            static=True,
        )
        return MapInfo(**data["info"])

    async def get_map_labels(self, map_id: int = 2, *, lang: str = None) -> List[MapNode]:
        """Get the map tree of all categories & lables of map points

        :param map_id: The id of the map
        :param lang: The language to use
        """
        data = await self.request_map("map/label/tree", lang=lang, map_id=map_id, static=True)
        return [MapNode(**i) for i in data["tree"]]

    async def get_map_points(
        self, map_id: int = 2, *, lang: str = None
    ) -> Tuple[List[MapNode], List[MapPoint]]:
        """Get a tuple of all map lables and map points

        :param map_id: The id of the map
        :param lang: The language to use
        """
        data = await self.request_map("map/point/list", lang=lang, map_id=map_id, static=True)

        labels = [MapNode(**i) for i in data["label_list"]]
        points = [MapPoint(**i) for i in data["point_list"]]

        return labels, points

    async def get_map_locations(self, map_id: int = 2, *, lang: str = None) -> List[MapLocation]:
        """Get a list of all locations on a map

        :param map_id: The id of the map
        :param lang: The language to use
        """
        data = await self.request_map("map/map_anchor/list", lang=lang, map_id=map_id, static=True)

        return [MapLocation(**i) for i in data["list"]]

    # MISC:

    async def _complete_uid(
        self,
        uid: Optional[int] = None,
        uid_key: str = "uid",
        server_key: str = "region",
    ) -> Dict[str, Any]:
        """Create a new dict with a uid and a server

        These are fetched from the currently authenticated user
        """
        params: Dict[str, Any] = {}

        uid = uid or self._uid

        if uid is None:
            accounts = await self.genshin_accounts()
            # filter test servers
            accounts = [
                account for account in accounts if "os" in account.server or "cn" in account.server
            ]

            # TODO: Raise properly
            if not accounts:
                errors.raise_for_retcode({"retcode": -1073})

            account = max(accounts, key=lambda a: a.level)
            uid = account.uid

            self._uid = uid

        params[uid_key] = uid
        params[server_key] = recognize_server(uid)

        return params

    async def _fetch_mi18n(self) -> Dict[str, Dict[str, str]]:
        if self.fetched_mi18n:
            return GenshinModel._mi18n

        self.fetched_mi18n = True

        async def single(url: str, key: str, lang: str = None):
            if lang is None:
                coros = (single(url, key, l) for l in LANGS)
                return await asyncio.gather(*coros)

            data = await self.request_webstatic(url.format(lang=lang))
            for k, v in data.items():
                GenshinModel._mi18n.setdefault(key + "/" + k, {})[lang] = v

        coros = (single(url, key) for key, url in GenshinModel._mi18n_urls.items())
        await asyncio.gather(*coros)

        return GenshinModel._mi18n

    async def init(self, lang: str = None):
        """Request all static & permanent endpoints to not require them later

        :param lang: The language to use
        """
        lang = lang or self.lang

        await asyncio.gather(
            self.get_banner_names(lang=lang),
            self.get_banner_details(lang=lang),
            self.get_monthly_rewards(),
            self._get_transaction_reasons(lang=lang),
            self._fetch_mi18n(),
        )

    async def fetch_banner_ids(self) -> List[str]:
        """Fetch banner ids from a user-mantained repo"""
        url = "https://raw.githubusercontent.com/thesadru/genshindata/master/banner_ids.txt"
        async with self.session.get(url) as r:
            data = await r.text()
        return data.splitlines()


class ChineseClient(GenshinClient):
    """A Genshin Client for chinese endpoints"""

    DS_SALT = "xV8v4Qu54lUKrEYFZkJhB8cuOh9Asafs"
    SIGNIN_SALT = "4a8knnbk5pbjqsrudp3dq484m9axoc5g"
    ACT_ID = "e202009291139501"

    TAKUMI_URL = "https://api-takumi.mihoyo.com/"
    RECORD_URL = "https://api-takumi-record.mihoyo.com/game_record/app/"
    INFO_LEDGER_URL = "https://hk4e-api.mihoyo.com/event/ys_ledger/monthInfo"
    DETAIL_LEDGER_URL = "https://hk4e-api.mihoyo.com/event/ys_ledger/monthDetail"
    REWARD_URL = "https://api-takumi.mihoyo.com/event/bbs_sign_reward/"
    GACHA_INFO_URL = "https://hk4e-api.mihoyo.com/event/gacha_info/api/"
    MAP_URL = "https://api-takumi-static.mihoyo.com/common/map_user/ys_obc/v1/map/"
    STATIC_MAP_URL = "https://api-static.mihoyo.com/common/map_user/ys_obc/v1/map"

    def __init__(
        self,
        cookies: Mapping[str, str] = None,
        authkey: str = None,
        *,
        lang: str = "zh-tw",
        debug: bool = False,
    ) -> None:
        super().__init__(cookies=cookies, authkey=authkey, lang=lang, debug=debug)

    async def request_hoyolab(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        json: Any = None,
        params: Dict[str, Any] = None,
        cache: Tuple[Any, ...] = None,
        cache_check: Callable[[Any], bool] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        kwargs.pop("lang", None)
        params = params or {}

        if cache:
            data = await self._check_cache(cache, cache_check)
            if data:
                return data

        if not self.cookies:
            raise RuntimeError("No cookies provided")

        url = URL(self.TAKUMI_URL).join(URL(endpoint))

        # all of this repetition is literally just to change these few lines
        headers = {
            "x-rpc-app_version": "2.11.1",
            "x-rpc-client_type": "5",
            "ds": generate_cn_dynamic_secret(self.DS_SALT, json, params),
        }

        data = await self.request(url, method, headers=headers, json=json, params=params, **kwargs)

        if cache:
            await self._update_cache(data, cache, cache_check)

        return data

    async def request_ledger(
        self,
        uid: int = None,
        detail: bool = False,
        *,
        month: int = None,
        params: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Make a request towards the ys ledger endpoint

        Traveler's diary related data
        """
        kwargs.pop("lang", None)
        params = params or {}

        url = URL(self.DETAIL_LEDGER_URL if detail else self.INFO_LEDGER_URL)

        params.update(await self._complete_uid(uid, "bind_uid", "bind_region"))
        params["month"] = month or datetime.now().month

        return await self.request(url, params=params)

    async def request_daily_reward(
        self,
        endpoint: str,
        uid: int = None,
        *,
        method: str = "GET",
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        kwargs.pop("lang", None)
        headers = headers or {}
        params = params or {}

        if not self.cookies:
            raise RuntimeError("No cookies provided")

        url = URL(self.REWARD_URL).join(URL(endpoint))

        params.update(await self._complete_uid(uid))

        headers["x-rpc-app_version"] = "2.10.1"
        headers["x-rpc-client_type"] = "5"
        headers["x-rpc-device_id"] = str(uuid.uuid4())
        headers["ds"] = generate_dynamic_secret(self.SIGNIN_SALT)

        params["act_id"] = self.ACT_ID

        return await self.request(url, method, params=params, headers=headers, **kwargs)

    async def get_reward_info(self, uid: int = None) -> DailyRewardInfo:
        """Get the daily reward info for the current user

        :param uid: Genshin uid of the currently logged-in user
        """
        data = await self.request_daily_reward("info", uid)
        return DailyRewardInfo(data["is_sign"], data["total_sign_day"])

    async def get_monthly_rewards(self) -> List[DailyReward]:
        """Get a list of all availible rewards for the current month"""
        # uid doesn't matter, this is a static resource
        func = perm_cache(
            ("cn_rewards", datetime.utcnow().month),
            self.request_daily_reward,
        )
        data = await func("home", uid=1)
        return [DailyReward(**i) for i in data["awards"]]

    def claimed_rewards(self, uid: int = None, *, limit: int = None) -> ChineseDailyRewardPaginator:
        """Get all claimed rewards for the current user

        NOTE: Languages are currently broken,
        the language is based off the language used to claim the reward

        :param limit: The maximum amount of rewards to get
        :param lang: The language to use - currently broken
        """
        return ChineseDailyRewardPaginator(self, uid, limit=limit)

    @overload
    async def claim_daily_reward(
        self, uid: int = None, *, reward: Literal[True] = ...
    ) -> DailyReward:
        ...

    @overload
    async def claim_daily_reward(self, uid: int = None, *, reward: Literal[False]) -> None:
        ...

    async def claim_daily_reward(
        self, uid: int = None, *, reward: bool = True
    ) -> Optional[DailyReward]:
        """Signs into hoyolab and claims the daily reward.

        :param uid: Genshin uid of the currently logged-in user
        :param reward: Whether to also fetch the claimed reward
        """
        await self.request_daily_reward("sign", uid, method="POST")

        if not reward:
            return None

        info, rewards = await asyncio.gather(
            self.get_reward_info(),
            self.get_monthly_rewards(),
        )
        return rewards[info.claimed_rewards - 1]


class MultiCookieClient(GenshinClient):
    """A Genshin Client which allows setting multiple cookies"""

    sessions: List[aiohttp.ClientSession]

    def __init__(
        self,
        cookie_list: Iterable[Mapping[str, str]] = None,
        *,
        lang: str = "en-us",
        debug: bool = False,
    ) -> None:
        self.sessions = []

        if cookie_list:
            self.set_cookies(cookie_list)

        super().__init__(lang=lang, debug=debug)

    @property
    def session(self) -> aiohttp.ClientSession:
        """The currently chosen session"""
        if not self.sessions:
            return aiohttp.ClientSession()

        return self.sessions[0]

    @property
    def cookies(self) -> List[Mapping[str, str]]:
        """A list of all cookies"""
        return [{m.key: m.value for m in s.cookie_jar} for s in self.sessions]

    @cookies.setter
    def cookies(self, cookies: List[Mapping[str, Any]]) -> None:
        self.set_cookies(cookies)

    def set_cookies(
        self,
        cookie_list: Union[Iterable[Union[Mapping[str, Any], str]], str],
        clear: bool = True,
    ) -> List[Mapping[str, str]]:
        """Set a list of cookies

        :param cookie_list: A list of cookies or a json file containing cookies
        :param clear: Whether to clear all of the previous cookies
        """
        if clear:
            self.sessions.clear()

        if isinstance(cookie_list, str):
            with open(cookie_list) as file:
                cookie_list = json_.load(file)

            if not isinstance(cookie_list, list):
                raise RuntimeError("Json file must contain a list of cookies")

        for cookies in cookie_list:
            session = aiohttp.ClientSession(cookies=SimpleCookie(cookies))
            self.sessions.append(session)

        return self.cookies

    def set_browser_cookies(self, *args: Any, **kwargs: Any) -> NoReturn:
        raise RuntimeError(f"{type(self).__name__} does not support browser cookies")

    async def close(self) -> None:
        """Close the underlying aiohttp sessions"""
        tasks = [
            asyncio.create_task(session.close()) for session in self.sessions if not session.closed
        ]
        if tasks:
            await asyncio.wait(tasks)

        await super().close()

    async def request(
        self,
        url: Union[str, URL],
        method: str = "GET",
        headers: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        headers = headers or {}
        headers["user-agent"] = self.USER_AGENT

        for session in self.sessions.copy():
            async with session.request(method, url, headers=headers, **kwargs) as r:
                r.raise_for_status()
                data = await r.json()

            if data["retcode"] == 0:
                return data["data"]

            try:
                errors.raise_for_retcode(data)
            except errors.TooManyRequests:
                # move the ratelimited session to the end to let the ratelimit wear off
                session = self.sessions.pop(0)
                self.sessions.append(session)

        # if we're here it means we used up all our sessions so we must handle that
        msg = "All cookies have hit their request limit of 30 accounts per day."
        raise errors.TooManyRequests({"retcode": 10101}, msg)

    async def request_daily_reward(self, *args: Any, **kwargs: Any) -> NoReturn:
        """Helper overwrite to prevent nasty bugs"""
        raise RuntimeError(f"{type(self).__name__} does not support daily reward endpoints")


class ChineseMultiCookieClient(MultiCookieClient, ChineseClient):
    """A Genshin Client for chinese endpoints which allows setting multiple cookies"""


class InternationalClient:
    """A client dummy which accepts both overseas and chinese cookies

    Currently in development and probably shouldn't be used.
    """

    os_client: MultiCookieClient
    cn_client: ChineseMultiCookieClient

    def __init__(
        self,
        os_client: MultiCookieClient = None,
        cn_client: ChineseMultiCookieClient = None,
        *,
        debug: bool = False,
    ) -> None:
        warnings.warn("Initialized an unstable InteractionClient type")

        self.os_client = os_client or MultiCookieClient(debug=debug)
        self.cn_client = cn_client or ChineseMultiCookieClient(debug=debug)

    async def close(self):
        await asyncio.gather(self.os_client.close(), self.cn_client.close())

    async def set_cookies(
        self,
        cookies: Sequence[Mapping[str, Any]] = None,
        *,
        os: Sequence[Mapping[str, Any]] = None,
        cn: Sequence[Mapping[str, Any]] = None,
    ) -> Tuple[List[Mapping[str, Any]], List[Mapping[str, Any]]]:
        """Helper cookie setter that accepts cookie headers

        It is recommended to set os and cn cookies explicitly.

        :param cookies: A list of either os or cn cookies if unknown
        :param os: A list of cookies known to be os
        :param cn: A list of cookies known to be cn
        :returns: A tuple of os cookies and cn cookies
        """
        cookies = cookies or []
        os = os or []
        cn = cn or []

        async def is_chinese(cookies: Mapping[str, Any]) -> bool:
            os = GenshinClient(cookies)
            cn = ChineseClient(cookies)

            osd, cnd = await asyncio.gather(
                os.request_hoyolab(f"community/user/wapi/getUserFullInfo?uid={os.hoyolab_uid}"),
                cn.request_hoyolab(f"user/wapi/getUserFullInfo?uid={os.hoyolab_uid}"),
                return_exceptions=True,
            )
            await asyncio.gather(os.close(), cn.close())

            if not isinstance(cnd, Exception):
                return True
            elif not isinstance(osd, Exception):
                return False

            raise Exception(f"Cookie {cookie!r} can't be found on neither os nor cn servers")

        self.os_client.set_cookies(os, clear=False)
        self.cn_client.set_cookies(cn, clear=False)

        are_chinese = await asyncio.gather(*(is_chinese(i) for i in cookies))
        for cookie, chinese in zip(cookies, are_chinese):
            if chinese:
                self.cn_client.set_cookies([cookie], clear=False)
            else:
                self.os_client.set_cookies([cookie], clear=False)

        return self.os_client.cookies, self.cn_client.cookies

    def get_activities(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_activities(uid, *args, **kwargs)

    def get_characters(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_characters(uid, *args, **kwargs)

    def get_full_user(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_full_user(uid, *args, **kwargs)

    def get_partial_user(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_partial_user(uid, *args, **kwargs)

    def get_spiral_abyss(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_spiral_abyss(uid, *args, **kwargs)

    def get_user(self, uid: int, *args: Any, **kwargs: Any):
        client = self.cn_client if is_chinese(uid) else self.os_client
        return client.get_user(uid, *args, **kwargs)

    async def get_record_card(self, hoyolab_uid: int, *args: Any, **kwargs: Any):
        try:
            return await self.os_client.get_record_card(hoyolab_uid, *args, **kwargs)
        except:
            return await self.cn_client.get_record_card(hoyolab_uid, *args, **kwargs)



"""A cli for genshin.py"""
import asyncio
from functools import update_wrapper
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, TypeVar

import difflib
import typer
import click

import genshin

T = TypeVar("T", bound=Any)

app = typer.Typer(name="genshin")


def asynchronous(func: Callable[..., Awaitable[Any]]) -> Callable[..., Any]:
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return update_wrapper(wrapper, func)


@app.command()
@asynchronous
async def accounts(
    lang: str = typer.Option("en-us", help="The language to use"),
):
    """Get all of your genshin accounts"""
    async with genshin.GenshinClient() as client:
        client.set_browser_cookies()
        data = await client.genshin_accounts(lang=lang)

    for account in data:
        typer.echo(
            f"{typer.style(str(account.uid), bold=True)} - {account.nickname} AR {account.level} ({account.server_name})"
        )


@app.command()
@asynchronous
async def stats(
    uid: int = typer.Argument(..., help="A genshin uid"),
    lang: str = typer.Option("en-us", help="The language to use"),
):
    """Show simple statistics of a user"""
    cuid = typer.style(str(uid), fg="blue")
    typer.echo(f"User stats of {cuid}\n")

    async with genshin.GenshinClient() as client:
        client.set_browser_cookies()
        data = await client.get_partial_user(uid, lang=lang)

    typer.secho("Stats:", fg="yellow")
    for k, v in data.stats.as_dict().items():
        value = typer.style(str(v), bold=True)
        typer.echo(f"{k}: {value}")

    typer.echo()
    typer.secho("Explorations:", fg="yellow")
    for area in data.explorations:
        perc = typer.style(str(area.percentage) + "%", bold=True)
        typer.echo(f"{area.name}: explored {perc} | {area.type} level {area.level}")

    if data.teapot is not None:
        typer.echo()
        typer.secho("Teapot:", fg="yellow")
        level = typer.style(str(data.teapot.level), bold=True)
        comfort = typer.style(str(data.teapot.comfort), bold=True)
        typer.echo(f"level {level} | comfort {comfort} ({data.teapot.comfort_name})")
        typer.echo(f"Unlocked realms: {', '.join(r.name for r in data.teapot.realms)}")


@app.command()
@asynchronous
async def characters(
    uid: int = typer.Argument(..., help="A genshin uid"),
    lang: str = typer.Option("en-us", help="The language to use"),
):
    """Show the characters of a user"""
    cuid = typer.style(str(uid), fg="blue")
    typer.echo(f"Characters of {cuid}")

    async with genshin.GenshinClient() as client:
        client.set_browser_cookies()
        data = await client.get_user(uid, lang=lang)

    characters = sorted(data.characters, key=lambda c: (c.level, c.rarity), reverse=True)
    for char in characters:
        color = {
            "Anemo": "bright_green",
            "Pyro": "red",
            "Hydro": "bright_blue",
            "Electro": "magenta",
            "Cryo": "bright_cyan",
            "Geo": "yellow",
            "Dendro": "green",
        }[char.element]

        typer.echo()
        name = typer.style(char.name, bold=True)
        element = typer.style(char.element, fg=color)
        typer.echo(f"{name} ({'★' * char.rarity} {element})")
        typer.echo(f"lvl {char.level} C{char.constellation}, friendship lvl {char.friendship}")
        typer.echo(
            f"Weapon: {char.weapon.name} ({'★' * char.weapon.rarity} {char.weapon.type}) - "
            f"lvl {char.weapon.level} R{char.weapon.refinement}"
        )
        if char.artifacts:
            typer.echo("Artifacts:")
            for arti in char.artifacts:
                typer.echo(f" - {arti.pos_name}: {arti.set.name} ({'★' * arti.rarity})")
        if char.outfits:
            typer.echo(f"Outfits: {', '.join(o.name for o in char.outfits)}")


@app.command()
@asynchronous
async def notes(
    uid: int = typer.Argument(..., help="A genshin uid"),
    lang: str = typer.Option("en-us", help="The language to use"),
):
    """Real-Time notes to see live genshin data"""
    cuid = typer.style(str(uid), fg="blue")
    typer.echo(f"Real-Time notes of {cuid}")

    async with genshin.GenshinClient() as client:
        client.set_browser_cookies()
        data = await client.get_notes(uid, lang=lang)

    typer.echo(f"{typer.style('Resin:', bold=True)} {data.current_resin}/{data.max_resin}")
    typer.echo(
        f"{typer.style('Comissions:', bold=True)} "
        f"{data.completed_commissions}/{data.max_comissions}",
        nl=False,
    )
    if data.completed_commissions == data.max_comissions and not data.claimed_comission_reward:
        typer.echo(f" | [{typer.style('X', fg='red')}] Haven't claimed rewards")
    else:
        typer.echo()
    typer.echo(
        f"{typer.style('Used resin cost-halving opportunities:', bold=True)} "
        f"{data.max_resin_discounts - data.remaining_resin_discounts}/{data.max_resin_discounts}"
    )

    typer.echo(
        f"\n{typer.style('Expeditions:', bold=True)} "
        f"{len(data.expeditions)}/{data.max_expeditions}"
    )
    for expedition in data.expeditions:
        if expedition.remaining_time > 0:
            seconds = expedition.remaining_time
            remaining = f"{seconds // 3600:02.0f}:{seconds % 3600 // 60:02.0f} remaining"
            typer.echo(f" - {expedition.status} | {remaining} - {expedition.character.name}")
        else:
            typer.echo(f" - {expedition.status} | {expedition.character.name}")


@app.command()
@asynchronous
async def wishes(
    limit: int = typer.Option(None, help="The maximum amount of wishes to show"),
    lang: str = typer.Option("en-us", help="The language to use"),
):
    """Get a nicely formatted wish history"""
    async with genshin.GenshinClient() as client:
        client.set_authkey()

        banner_names = await client.get_banner_names(lang=lang)
        longest = max(len(v) for v in banner_names.values())

        async for wish in client.wish_history(limit=limit, lang=lang):
            banner = typer.style(wish.banner_name.ljust(longest), bold=True)
            typer.echo(f"{banner} | {wish.time} - {wish.name} ({'★' * wish.rarity} {wish.type})")


@app.command()
@asynchronous
async def pity():
    """Calculates the amount of pulls until pity"""
    async with genshin.GenshinClient() as client:
        client.set_authkey()

        banners = await client.get_banner_names()
        for banner, name in banners.items():
            # skip the novice banner
            if banner == 100:
                continue

            typer.echo()
            typer.secho(name, fg="yellow")

            accum = 0
            async for wish in client.wish_history(banner):
                accum += 1
                if wish.rarity == 5:
                    a = typer.style(str(90 - accum), bold=True)
                    n = typer.style(wish.name, fg="green")
                    typer.secho(f"Pulled a {n} {accum} pulls ago. {a} pulls left until pity.")
                    break
            else:
                a = typer.style(str(90 - accum), bold=True)
                typer.secho(f"Never pulled a 5*. At most {a} pulls left until pity")


@app.command()
@asynchronous
async def calculate():
    """Calcualte the amount of resources needed to upgrade anything"""
    async with genshin.GenshinClient() as client:
        client.set_browser_cookies()

        characters = await client.get_calculator_characters()

        inp: Any = typer.prompt(f"Character (name or id)")

        names = {obj.name: obj for obj in characters}
        matches = difflib.get_close_matches(inp, names, n=1)
        if matches:
            character = names[matches[0]]
        elif not inp.isdigit():
            raise click.ClickException("Invalid character name")
        else:
            inp = int(inp)
            for character in characters:
                if character.id == inp:
                    break

            raise click.ClickException("Invalid character format")

        current: Any = typer.prompt("Current level", type=int)
        target: Any = typer.prompt("Target level", type=int)
        result = await client.calculator().set_character(character, current=current, target=target)

        typer.echo()
        typer.echo(f"Items needed to level up {character.name} from lvl {current} to {target}:")

        for item in result.total:
            typer.echo(f"{item.amount}x {item.name}")


@app.command()
@asynchronous
async def banner_ids():
    """Get the banner ids from logs"""
    ids = genshin.get_banner_ids()

    typer.echo("\n".join(ids) + "\n")

    if len(ids) < 3:
        typer.echo("Please open all detail pages!")
        return

    async with genshin.GenshinClient() as client:
        for banner in await client.get_banner_details(ids):
            typer.echo(f"{banner.banner_id} - {banner.banner_type_name} ({banner.banner_type})")


if __name__ == "__main__":
    app()



"""Errors raised by genshin.py"""
from typing import Any, Dict, NoReturn, Optional, Tuple, Type, Union

__all__ = [
    "GenshinException",
    # game record:
    "AccountNotFound",
    "DataNotPublic",
    # cookies:
    "CookieException",
    "InvalidCookies",
    "TooManyRequests",
    # daily:
    "AlreadyClaimed",
    # redemption:
    "RedemptionInvalid",
    "RedemptionClaimed",
    "RedemptionCooldown",
    "RedemptionException",
    # authkeys:
    "AuthkeyException",
    "InvalidAuthkey",
    "AuthkeyTimeout",
    # misc:
    "ERRORS",
    "raise_for_retcode",
]


class GenshinException(Exception):
    """A base genshin exception"""

    retcode: int
    original: str
    msg: str = ""

    def __init__(self, response: Dict[str, Any], msg: str = None) -> None:
        self.retcode = response.get("retcode", 0)
        self.original = response.get("message", "")
        self.msg = msg or self.msg or self.original

        if self.retcode:
            msg = f"[{self.retcode}] {self.msg}"

        super().__init__(msg)

    def __repr__(self) -> str:
        response = {"retcode": self.retcode, "message": self.original}
        args = [repr(response)]
        if self.msg != self.original:
            args.append(repr(self.msg))

        return f"{type(self).__name__}({', '.join(args)})"

    @property
    def response(self) -> Dict[str, Any]:
        return {"retcode": self.retcode, "message": self.original, "data": None}


class AccountNotFound(GenshinException):
    """Tried to get data with an invalid uid."""

    msg = "Could not find user; uid may be invalid."


class DataNotPublic(GenshinException):
    """User hasn't set their data to public."""

    msg = "User's data is not public"


class CookieException(GenshinException):
    """Base error for cookies"""


class InvalidCookies(CookieException):
    """Cookies weren't valid"""

    msg = "Cookies are not valid"


class TooManyRequests(CookieException):
    """Made too many requests and got ratelimited"""

    msg = "Cannnot get data for more than 30 accounts per cookie per day."


class VisitsTooFrequently(GenshinException):
    """Visited a page too frequently, just handle with exponantial backoff"""

    msg = "visits too frequently"


class AlreadyClaimed(GenshinException):
    """Already claimed the daily reward today"""

    msg = "Already claimed the daily reward today."


class AuthkeyException(GenshinException):
    """Base error for authkeys"""


class InvalidAuthkey(AuthkeyException):
    """Authkey is not valid"""

    msg = "Authkey is not valid."


class AuthkeyTimeout(AuthkeyException):
    """Authkey has timed out"""

    msg = "Authkey has timed out."


class RedemptionException(GenshinException):
    """Exception caused by redeeming a code"""


class RedemptionInvalid(RedemptionException):
    """Invalid redemption code."""

    msg = "Invalid redemption code."


class RedemptionCooldown(RedemptionException):
    """Cooldown for redeeming"""

    msg = "Redemption is on cooldown."


class RedemptionClaimed(RedemptionException):
    """Redeption code has been claimed already."""

    msg = "Code has been claimed already."


_TGE = Type[GenshinException]
_errors: Dict[int, Union[_TGE, str, Tuple[_TGE, Optional[str]]]] = {
    # misc hoyolab
    -100: InvalidCookies,
    -108: "Invalid language.",
    -110: VisitsTooFrequently,
    # game record
    10001: InvalidCookies,
    -10001: "Malformed request.",
    -10002: "No genshin account associated with cookies.",
    # database game record
    10101: TooManyRequests,
    10102: DataNotPublic,
    10103: (InvalidCookies, "Cookies are valid but do not have a hoyolab account bound to them."),
    10104: "Tried to use a beta feature in an invalid context",
    # calculator
    -500001: "Invalid fields in calculation",
    -502001: "User does not have this character",
    # mixin
    -1: "Internal database error.",
    1009: AccountNotFound,
    # redemption
    -1071: InvalidCookies,
    -1073: (AccountNotFound, "Account has no game account bound to it."),
    -2001: (RedemptionInvalid, "Redemption code has expired."),
    -2004: RedemptionInvalid,
    -2016: (RedemptionCooldown, "Redemption in cooldown."),
    -2017: (RedemptionClaimed, "Redeption code has been claimed already."),
    -2021: (
        RedemptionException,
        "Cannot claim codes for accounts with adventure rank lower than 10.",
    ),
    # rewards
    -5003: AlreadyClaimed,
    # chinese
    1008: AccountNotFound,
    -1104: "This action must be done in the app",
}

ERRORS: Dict[int, Tuple[Type[GenshinException], Optional[str]]] = {
    retcode: (
        (exc, None)
        if isinstance(exc, type)
        else (GenshinException, exc)
        if isinstance(exc, str)
        else exc
    )
    for retcode, exc in _errors.items()
}


def raise_for_retcode(data: Dict[str, Any]) -> NoReturn:
    """Raise an equivalent error to a response

    game record:
        10001 = invalid cookie
        101xx = generic errors

    authkey:
        -100 = invalid authkey
        -101 = authkey timed out

    code redemption:
        20xx = invalid code or state
        -107x = invalid cookies

    daily reward:
        -500x = already claimed the daily reward
    """
    r, m = data.get("retcode", 0), data.get("message", "")

    if m.startswith("authkey"):
        if r == -100:
            raise InvalidAuthkey(data)
        elif r == -101:
            raise AuthkeyTimeout(data)
        else:
            raise AuthkeyException(data)

    elif m.startswith("character id"):
        char = m.split(":")[1].split()[0]
        raise GenshinException(data, f"User does not have a character with id {char}")

    elif r in ERRORS:
        exctype, msg = ERRORS[r]
        raise exctype(data, msg)

    else:
        raise GenshinException(data)



"""Modern API wrapper for Genshin Impact built on asyncio and pydantic.

Documentation: https://thesadru.github.io/genshin.py

Source Code: https://github.com/thesadru/genshin.py
"""
from . import models
from .client import *
from .constants import *
from .errors import *
from .utils import *



import re
from datetime import datetime
from typing import Any, List, Literal, Optional

from pydantic import Field, validator

from .base import GenshinModel, Unique

__all__ = [
    "BannerType",
    "BANNER_TYPES",
    "Wish",
    "BannerDetailItem",
    "BannerDetailsUpItem",
    "BannerDetails",
    "GachaItem",
]

BannerType = Literal[100, 200, 301, 302, 400]

BANNER_TYPES: List[BannerType] = [100, 200, 301, 302, 400]


class Wish(GenshinModel, Unique):
    """A wish made on any banner"""

    uid: int

    id: int
    type: str = Field(galias="item_type")
    name: str
    rarity: int = Field(galias="rank_type")
    time: datetime

    banner_type: BannerType = Field(galias="gacha_type")
    banner_name: str

    @validator("banner_type", pre=True)
    def __cast_banner_type(cls, v: Any) -> int:
        return int(v)


class BannerDetailItem(GenshinModel):
    """An item that may be gotten from a banner"""

    name: str = Field(galias="item_name")
    type: str = Field(galias="item_type")
    rarity: int = Field(galias="rank")

    up: bool = Field(galias="is_up")
    order: int = Field(galias="order_value")


class BannerDetailsUpItem(GenshinModel):
    """An item that has a rate-up on a banner"""

    name: str = Field(galias="item_name")
    type: str = Field(galias="item_type")
    element: str = Field(galias="item_attr")
    icon: str = Field(galias="item_img")

    @validator("element", pre=True)
    def __parse_element(cls, v: str) -> str:
        return {
            "风": "Anemo",
            "火": "Pyro",
            "水": "Hydro",
            "雷": "Electro",
            "冰": "Cryo",
            "岩": "Geo",
            "草": "Dendro",
            "": "",
        }[v]


class BannerDetails(GenshinModel):
    """Details of a banner"""

    banner_id: str
    banner_type: int = Field(galias="gacha_type")
    title: str
    content: str
    date_range: str

    r5_up_prob: Optional[float]
    r4_up_prob: Optional[float]
    r5_prob: Optional[float]
    r4_prob: Optional[float]
    r3_prob: Optional[float]
    r5_guarantee_prob: Optional[float] = Field(galias="r5_baodi_prob")
    r4_guarantee_prob: Optional[float] = Field(galias="r4_baodi_prob")
    r3_guarantee_prob: Optional[float] = Field(galias="r3_baodi_prob")

    r5_up_items: List[BannerDetailsUpItem]
    r4_up_items: List[BannerDetailsUpItem]

    r5_items: List[BannerDetailItem] = Field(galias="r5_prob_list")
    r4_items: List[BannerDetailItem] = Field(galias="r4_prob_list")
    r3_items: List[BannerDetailItem] = Field(galias="r3_prob_list")

    @validator("r5_up_items", "r4_up_items", pre=True)
    def __replace_none(cls, v: Optional[List[Any]]) -> List[Any]:
        return v or []

    @validator(
        "r5_up_prob",
        "r4_up_prob",
        "r5_prob",
        "r4_prob",
        "r3_prob",
        "r5_guarantee_prob",
        "r4_guarantee_prob",
        "r3_guarantee_prob",
        pre=True,
    )
    def __parse_percentage(cls, v: str) -> Optional[float]:
        return None if v == "0%" else float(v[:-1].replace(",", "."))

    @property
    def name(self) -> str:
        return re.sub(r"<.*?>", "", self.title).strip()

    @property
    def banner_type_name(self) -> str:
        banners = {
            100: "Novice Wishes",
            200: "Permanent Wish",
            301: "Character Event Wish",
            302: "Weapon Event Wish",
            400: "Character Event Wish",
        }
        return banners[self.banner_type]

    @property
    def items(self) -> List[BannerDetailItem]:
        items = self.r5_items + self.r4_items + self.r3_items
        return sorted(items, key=lambda x: x.order)


class GachaItem(GenshinModel, Unique):
    """An item that can be gotten from the gacha"""

    name: str
    type: str = Field(galias="item_type")
    rarity: int = Field(galias="rank_type")
    id: int = Field(galias="item_id")

    @validator("id")
    def __format_id(cls, v: int) -> int:
        return 10000000 + v - 1000 if len(str(v)) == 4 else v

    def is_character(self) -> bool:
        return len(str(self.id)) == 8



"""Utility functions related to genshin fuckery"""
import hashlib
import json
import random
import re
import string
import time
from typing import Any, Mapping, Union

__all__ = [
    "generate_dynamic_secret",
    "generate_cn_dynamic_secret",
    "create_short_lang_code",
    "recognize_server",
    "recognize_id",
    "is_genshin_uid",
    "is_chinese",
]


def generate_dynamic_secret(salt: str) -> str:
    """Creates a new overseas dynamic secret

    :param salt: A ds salt
    """
    t = int(time.time())
    r = "".join(random.choices(string.ascii_letters, k=6))
    h = hashlib.md5(f"salt={salt}&t={t}&r={r}".encode()).hexdigest()
    return f"{t},{r},{h}"


def generate_cn_dynamic_secret(salt: str, body: Any = None, query: Mapping[str, Any] = None) -> str:
    """Creates a new chinese dynamic secret

    :param salt: A ds salt
    :param body: A json body of the request
    :param query: A query of the request
    """
    t = int(time.time())
    r = random.randint(100001, 200000)
    b = json.dumps(body) if body else ""
    q = "&".join(f"{k}={v}" for k, v in sorted(query.items())) if query else ""

    h = hashlib.md5(f"salt={salt}&t={t}&r={r}&b={b}&q={q}".encode()).hexdigest()
    return f"{t},{r},{h}"


def create_short_lang_code(lang: str) -> str:
    """Returns an alternative short lang code

    :param lang: A language code
    """
    return lang if "zh" in lang else lang.split("-")[0]


def recognize_server(uid: int) -> str:
    """Recognizes which server a UID is from.

    :param uid: A genshin uid
    """
    server = {
        "1": "cn_gf01",
        "2": "cn_gf01",
        "5": "cn_qd01",
        "6": "os_usa",
        "7": "os_euro",
        "8": "os_asia",
        "9": "os_cht",
    }.get(str(uid)[0])

    if server:
        return server
    else:
        raise ValueError(f"UID {uid} isn't associated with any server")


def recognize_id(id: int):
    """Attempts to recognize what item type an id is

    :param id: Any id
    :returns: A model for the appropriate id
    """
    from .. import models

    if 10000000 < id < 20000000:
        return models.base.BaseCharacter
    elif 1000000 < id < 10000000:
        return models.character.ArtifactSet
    elif 100000 < id < 1000000:
        return models.character.Outfit
    elif 50000 < id < 100000:
        return models.character.Artifact
    elif 10000 < id < 50000:
        return models.character.Weapon
    elif 100 < id < 1000:
        return models.character.Constellation
    elif 10e17 < id < 10e19:
        return models.wish.Wish
    else:
        return None


def is_genshin_uid(uid: int) -> bool:
    """Recognizes whether the uid is a valid genshin uid

    Doesn't work for chinese uids.

    :param uid: A genshin or hoyolab uid
    """
    return bool(re.fullmatch(r"[6789]\d{8}", str(uid)))


def is_chinese(x: Union[int, str]) -> bool:
    """Recognizes whether the server/uid is chinese

    :param x: A server or a genshin uid
    """
    return str(x).startswith(("cn", "1", "2", "5"))



"""Utilities for iterators"""
import asyncio
import heapq
from typing import *

T = TypeVar("T")

__all__ = ["aenumerate", "aislice", "azip", "amerge"]


async def aenumerate(iterable: AsyncIterable[T], start: int = 0) -> AsyncIterator[Tuple[int, T]]:
    """enumerate for async iterators

    :param iterable: An async iterable
    :param start: The starting value
    """
    __tracebackhide__ = True  # for pytest

    i = start
    async for x in iterable:
        yield i, x
        i += 1


async def aislice(iterable: AsyncIterable[T], stop: int = None) -> AsyncIterator[T]:
    """islice for async iterators

    :param iterable: An async iterable
    :param stop: The stopping index, if any
    """
    __tracebackhide__ = True  # for pytest

    async for i, x in aenumerate(iterable, start=1):
        yield x

        if stop and i >= stop:
            return


async def azip(*iterables: AsyncIterable[T]) -> AsyncIterator[Tuple[T, ...]]:
    """zip for async iterators

    :param iterables: Async iterables
    """
    iterators = [i.__aiter__() for i in iterables]
    while True:
        coros = (i.__anext__() for i in iterators)
        try:
            x = await asyncio.gather(*coros)
        except StopAsyncIteration:
            break
        yield tuple(x)


async def _try_gather_first(iterators: Iterable[AsyncIterator[T]]) -> List[T]:
    """Gather all the first values of iterators at once

    :param iterators: Async iterators
    """
    __tracebackhide__ = True  # for pytest

    coros = (it.__anext__() for it in iterators)
    gathered = await asyncio.gather(*coros, return_exceptions=True)
    values = []
    for x in gathered:
        if isinstance(x, BaseException):
            if not isinstance(x, StopAsyncIteration):
                raise x from None
        else:
            values.append(x)
    return values


async def amerge(
    iterables: Iterable[AsyncIterable[T]], key: Callable[[T], Any] = None
) -> AsyncIterator[T]:
    """heapq.merge for async iterators

    :param iterables: Async iterables
    :param key: Function to determine the sort order
    """
    __tracebackhide__ = True  # for pytest

    key = key or (lambda x: x)

    iterators = [i.__aiter__() for i in iterables]
    values = await _try_gather_first(iterators)

    heap: List[List[Any]] = [
        [key(value), order, value, it.__anext__]
        for order, (it, value) in enumerate(zip(iterators, values))
    ]

    # the rest is simply heapq.merge:
    heapq.heapify(heap)

    while len(heap) > 1:
        try:
            while True:
                _, _, value, anext = s = heap[0]
                yield value
                value = await anext()
                s[0] = key(value)
                s[2] = value
                heapq.heapreplace(heap, s)
        except StopAsyncIteration:
            heapq.heappop(heap)

    if heap:
        _, _, value, anext = heap[0]
        yield value
        async for item in anext.__self__:
            yield item



from datetime import datetime
from typing import Literal

from pydantic import Field

from .base import GenshinModel, Unique

TransactionKind = Literal["primogem", "crystal", "resin", "artifact", "weapon"]

__all__ = ["TransactionKind", "BaseTransaction", "Transaction", "ItemTransaction"]


class BaseTransaction(GenshinModel, Unique):
    """A genshin transaction"""

    kind: TransactionKind

    id: int
    uid: int
    time: datetime
    amount: int = Field(galias="add_num")
    reason_id: int
    reason: str = ""


class Transaction(BaseTransaction):
    """A genshin transaction of currency"""

    kind: Literal["primogem", "crystal", "resin"]


class ItemTransaction(BaseTransaction):
    """A genshin transaction of artifacts or weapons"""

    kind: Literal["artifact", "weapon"]

    name: str
    rarity: int = Field(galias="rank")



"""Miscalenious utilities"""
import asyncio
import functools
import os
import tempfile
from asyncio.proactor_events import _ProactorBasePipeTransport
from typing import Any, Callable, Coroutine, Dict, Type, TypeVar

from .. import errors

CallableT = TypeVar("CallableT", bound=Callable[..., Coroutine[Any, Any, Any]])

__all__ = ["get_browser_cookies", "get_tempdir", "handle_ratelimits"]


def get_browser_cookies(browser: str = None) -> Dict[str, str]:
    """Gets cookies from your browser for later storing.

    Avalible browsers: chrome, chromium, opera, edge, firefox

    :param browser: The browser to extract the cookies from
    :returns: The extracted cookies
    """
    if browser and browser not in ("chrome", "chromium", "opera", "edge", "firefox"):
        raise ValueError(f"Unsupported browser: {browser}")

    import browser_cookie3

    load = getattr(browser_cookie3, browser.lower()) if browser else browser_cookie3.load

    allowed_cookies = {"ltuid", "ltoken", "account_id", "cookie_token"}
    return {
        c.name: c.value
        for domain in ("mihoyo", "hoyolab")
        for c in load(domain_name=domain)
        if c.name in allowed_cookies and c.value is not None
    }


def get_tempdir():
    """Gets the temporary directory to be used by genshin.py"""
    directory = os.path.join(tempfile.gettempdir(), "genshin.py")
    os.makedirs(directory, exist_ok=True)
    return directory


def handle_ratelimits(
    tries: int = 5,
    exception: Type[errors.GenshinException] = errors.VisitsTooFrequently,
    delay: float = 0.3,
) -> Callable[[CallableT], CallableT]:
    """Handles ratelimits for requests"""
    # TODO: Support exponential backoff

    def wrapper(func) -> Any:
        async def inner(self, *args: Any, **kwargs: Any) -> Any:
            for _ in range(tries):
                try:
                    x = await func(self, *args, **kwargs)
                except exception:
                    await asyncio.sleep(delay)
                else:
                    return x
            else:
                raise exception({}, f"Got ratelimited {tries} times in a row")

        return functools.update_wrapper(inner, func)

    return wrapper


# for the convenience of the user we hide these windows errors:
def _silence_event_loop_closed(func):
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except RuntimeError as e:
            if str(e) != "Event loop is closed":
                raise

    return wrapper


_ProactorBasePipeTransport.__del__ = _silence_event_loop_closed(_ProactorBasePipeTransport.__del__)  # type: ignore # noqa



from __future__ import annotations

import abc
import re
import warnings
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Union

from pydantic import BaseModel, Field, root_validator
from pydantic.fields import ModelField

from ..constants import CHARACTER_NAMES, DBChar

__all__ = [
    "GenshinModel",
    "Unique",
    "CharacterIcon",
    "BaseCharacter",
    "PartialCharacter",
]


class GenshinModel(BaseModel, abc.ABC):
    """A genshin model"""

    # nasty pydantic bug fixed only on the master branch - waiting for pypi release
    if TYPE_CHECKING:
        _mi18n_urls: ClassVar[Dict[str, str]]
        _mi18n: ClassVar[Dict[str, Dict[str, str]]]
    else:
        _mi18n_urls = {
            "bbs": "https://webstatic-sea.mihoyo.com/admin/mi18n/bbs_cn/m11241040191111/m11241040191111-{lang}.json",
        }
        _mi18n = {}

    def __init__(self, **data: Any) -> None:
        """"""
        # clear the docstring for pdoc
        super().__init__(**data)

    @root_validator(pre=True)
    def __parse_galias(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Due to alias being reserved for actual aliases we use a custom alias"""
        aliases = {}
        for name, field in cls.__fields__.items():
            alias = field.field_info.extra.get("galias")
            if alias is not None:
                aliases[alias] = name

        return {aliases.get(name, name): value for name, value in values.items()}

    @root_validator()
    def __parse_timezones(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        for name, field in cls.__fields__.items():
            if isinstance(values.get(name), datetime) and values[name].tzinfo is None:
                offset = field.field_info.extra.get("timezone", 0)
                tzinfo = timezone(timedelta(hours=offset))
                values[name] = values[name].replace(tzinfo=tzinfo)

        return values

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        """Generate a dictionary representation of the model,
        optionally specifying which fields to include or exclude.

        Takes the liberty of also giving properties as fields.
        """
        for name in dir(type(self)):
            clsvalue = getattr(type(self), name)
            if isinstance(clsvalue, property):
                try:
                    value = getattr(self, name)
                except:
                    continue

                if value == "":
                    continue

                self.__dict__[name] = value

        return super().dict(**kwargs)

    def _get_mi18n(self, field: ModelField, lang: str) -> str:
        key = field.field_info.extra.get("mi18n")
        if key not in self._mi18n:
            return field.name

        return self._mi18n[key][lang]

    class Config:
        allow_mutation = False


class Unique(abc.ABC):
    """A hashable model with an id"""

    id: int

    def __int__(self) -> int:
        return self.id

    def __hash__(self) -> int:
        return hash(self.id)


class CharacterIcon:
    """A character's icon"""

    character_name: str

    def __init__(self, icon: Union[str, int]) -> None:
        if isinstance(icon, int):
            char = CHARACTER_NAMES.get(icon)
            if char is None:
                raise ValueError(f"Invalid character id {icon}")
            self.character_name = char.icon_name
        else:
            match = re.search(r"UI_AvatarIcon(?:_Side)?_(.*).png", icon)
            self.character_name = icon if match is None else match.group(1)

    def create_icon(self, specifier: str, scale: int = 0) -> str:
        base = "https://upload-os-bbs.mihoyo.com/game_record/genshin/"
        return base + f"{specifier}_{self.character_name}{f'@{scale}x' if scale else ''}.png"

    @property
    def icon(self) -> str:
        return self.create_icon("character_icon/UI_AvatarIcon")

    @property
    def image(self) -> str:
        return self.create_icon("character_image/UI_AvatarIcon", scale=2)

    @property
    def side_icon(self) -> str:
        return self.create_icon("character_side_icon/UI_AvatarIcon_Side")

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.character_name!r})"


class BaseCharacter(GenshinModel, Unique):
    """A Base character model"""

    id: int
    name: str
    element: str
    rarity: int
    icon: str

    collab: bool = False

    @root_validator(pre=True)
    def __autocomplete(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Complete missing data"""
        partial: bool = False

        id = values.get("id")
        icon = values.get("icon") or values.get("image")
        name = values.get("name")

        if id and id in CHARACTER_NAMES:
            char = CHARACTER_NAMES[id]

        elif icon and "genshin" in icon:
            icon = CharacterIcon(icon)
            for char in CHARACTER_NAMES.values():
                if char.icon_name == icon.character_name:
                    break
            else:
                warnings.warn(f"Completing data for an unknown character\n{values!r}")
                name = icon.character_name
                char = DBChar(0, name, name, "Anemo", 5)

        elif name:
            for char in CHARACTER_NAMES.values():
                if char.name == name:
                    break
            else:
                warnings.warn(f"Completing data for a partial character\n{values!r}")
                partial = True
                char = DBChar(0, name, name, "Anemo", 5)

        else:
            raise ValueError("Character data incomplete")

        values["id"] = values.get("id") or char.id

        # malformed icon handling (in calculation)
        if "genshin" in (values.get("icon") or ""):
            values["icon"] = CharacterIcon(values["icon"]).icon
        elif values.get("icon"):
            if not partial:
                values["icon"] = CharacterIcon(char.icon_name).icon
        else:
            values["icon"] = CharacterIcon(char.icon_name).icon

        values["name"] = values.get("name") or char.name
        values["element"] = values.get("element") or char.element
        values["rarity"] = values.get("rarity") or char.rarity

        if values["rarity"] > 100:
            values["rarity"] -= 100
            values["collab"] = True
        elif values["id"] == 10000062:
            # sometimes Alot has 5* like normal, like cmon
            values["collab"] = True

        return values

    @property
    def image(self) -> str:
        return CharacterIcon(self.icon).image

    @property
    def side_icon(self) -> str:
        return CharacterIcon(self.icon).side_icon

    @property
    def traveler_name(self) -> str:
        if self.id == 10000005:
            return "Aether"
        elif self.id == 10000007:
            return "Lumine"
        else:
            return ""


class PartialCharacter(BaseCharacter):
    """A character without any equipment"""

    level: int
    friendship: int = Field(galias="fetter")
    constellation: int = Field(galias="actived_constellation_num")



from datetime import datetime
from typing import List

from pydantic import Field

from .base import GenshinModel

__all__ = [
    "BaseDiary",
    "DiaryActionCategory",
    "DiaryData",
    "DayDiaryData",
    "Diary",
    "DiaryAction",
    "DiaryPage",
]


class BaseDiary(GenshinModel):
    """Base model for diary and diary page"""

    uid: int
    region: str
    nickname: str
    month: int = Field(galias="data_month")


class DiaryActionCategory(GenshinModel):
    """A diary category for primogems"""

    id: int = Field(galias="action_id")
    name: str = Field(galias="action")
    amount: int = Field(galias="num")
    percentage: int = Field(galias="percent")


class DiaryData(GenshinModel):
    """Diary data for a month"""

    current_primogems: int
    current_mora: int
    last_primogems: int
    last_mora: int
    primogems_rate: int = Field(galias="primogem_rate")
    mora_rate: int
    categories: List[DiaryActionCategory] = Field(galias="group_by")


class DayDiaryData(GenshinModel):
    """Diary data for a day"""

    current_primogems: int
    current_mora: int


class Diary(BaseDiary):
    """A traveler's diary"""

    data: DiaryData = Field(galias="month_data")
    day_data: DayDiaryData


class DiaryAction(GenshinModel):
    """An action which earned currency"""

    action_id: int
    action: str
    time: datetime = Field(timezone=8)
    amount: int = Field(galias="num")


class DiaryPage(BaseDiary):
    """A page of a diary"""

    actions: List[DiaryAction] = Field(galias="list")



from __future__ import annotations

import collections
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from pydantic import Field, validator

from .base import BaseCharacter, GenshinModel, Unique

if TYPE_CHECKING:
    from genshin import GenshinClient

__all__ = [
    "CALCULATOR_ELEMENTS",
    "CALCULATOR_WEAPON_TYPES",
    "CALCULATOR_ARTIFACTS",
    "CalculatorCharacter",
    "CalculatorWeapon",
    "CalculatorArtifact",
    "CalculatorTalent",
    "CalculatorCharacterDetails",
    "CalculatorConsumable",
    "CalculatorArtifactResult",
    "CalculatorResult",
    "CalculatorObject",
    "CalculatorBuilder",
]

CALCULATOR_ELEMENTS: Dict[int, str] = {
    1: "Pyro",
    2: "Anemo",
    3: "Geo",
    4: "Dendro",
    5: "Electro",
    6: "Hydro",
    7: "Cryo",
}
CALCULATOR_WEAPON_TYPES: Dict[int, str] = {
    1: "Sword",
    10: "Catalyst",
    11: "Claymore",
    12: "Bow",
    13: "Polearm",
}
CALCULATOR_ARTIFACTS: Dict[int, str] = {
    1: "Flower of Life",
    2: "Plume of Death",
    3: "Sands of Eon",
    4: "Goblet of Eonothem",
    5: "Circlet of Logos",
}


class CalculatorCharacter(BaseCharacter):
    """A character meant to be used with calculators"""

    rarity: int = Field(galias="avatar_level")
    element: str = Field(galias="element_attr_id")
    weapon_type: str = Field(galias="weapon_cat_id")
    level: int = Field(0, galias="level_current")
    max_level: int

    @validator("element", pre=True)
    def __parse_element(cls, v: Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_ELEMENTS[int(v)]

    @validator("weapon_type", pre=True)
    def __parse_weapon_type(cls, v: Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_WEAPON_TYPES[int(v)]


class CalculatorWeapon(GenshinModel, Unique):
    """A weapon meant to be used with calculators"""

    id: int
    name: str
    icon: str
    rarity: int = Field(galias="weapon_level")
    type: str = Field(galias="weapon_cat_id")
    level: int = Field(0, galias="level_current")
    max_level: int

    @validator("type", pre=True)
    def __parse_weapon_type(cls, v: Any) -> str:
        if isinstance(v, str):
            return v

        return CALCULATOR_WEAPON_TYPES[int(v)]


class CalculatorArtifact(GenshinModel, Unique):
    """An artifact meant to be used with calculators"""

    id: int
    name: str
    icon: str
    rarity: int = Field(galias="reliquary_level")
    pos: int = Field(galias="reliquary_cat_id")
    level: int = Field(0, galias="level_current")
    max_level: int

    @property
    def pos_name(self) -> str:
        return CALCULATOR_ARTIFACTS[self.pos]


class CalculatorTalent(GenshinModel, Unique):
    """A talent of a character meant to be used with calculators"""

    id: int
    group_id: int
    name: str
    icon: str
    level: int = Field(0, galias="level_current")
    max_level: int

    @property
    def type(self) -> Literal["attack", "skill", "burst", "passive", "dash"]:
        """The type of the talent, parsed from the group id"""
        # It's Possible to parse this from the id too but group id feels more reliable

        # 4139 -> group=41 identifier=3 order=9
        group, relevant = divmod(self.group_id, 100)
        identifier, order = divmod(relevant, 10)

        if identifier == 2:
            return "passive"
        elif order == 1:
            return "attack"
        elif order == 2:
            return "skill"
        elif order == 9:
            return "burst"
        elif order == 3:
            return "dash"
        else:
            raise ValueError(f"Cannot parse type for talent {self.group_id!r} (group {group})")

    @property
    def upgradeable(self) -> bool:
        """Whether this talent can be leveled up"""
        return self.type not in ("passive", "dash")

    def __int__(self) -> int:
        return self.group_id


class CalculatorConsumable(GenshinModel, Unique):
    """An item consumed when upgrading"""

    id: int
    name: str
    icon: str
    amount: int = Field(galias="num")


class CalculatorCharacterDetails(GenshinModel):
    """Details of a synced calculator character"""

    weapon: CalculatorWeapon = Field(galias="weapon")
    talents: List[CalculatorTalent] = Field(galias="skill_list")
    artifacts: List[CalculatorArtifact] = Field(galias="reliquary_list")

    @validator("talents")
    def __correct_talent_current_level(cls, v: List[CalculatorTalent]) -> List[CalculatorTalent]:
        # passive talent have current levels at 0 for some reason
        talents: List[CalculatorTalent] = []

        for talent in v:
            if talent.max_level == 1 and talent.level == 0:
                raw = talent.dict()
                raw["level"] = 1
                talent = CalculatorTalent(**raw)

            talents.append(talent)

        return v


class CalculatorArtifactResult(GenshinModel):
    """A calulation result for a specific artifact"""

    artifact_id: int = Field(galias="reliquary_id")
    list: List[CalculatorConsumable] = Field(galias="id_consume_list")


class CalculatorResult(GenshinModel):
    """A calculation result"""

    character: List[CalculatorConsumable] = Field(galias="avatar_consume")
    weapon: List[CalculatorConsumable] = Field(galias="weapon_consume")
    talents: List[CalculatorConsumable] = Field(galias="avatar_skill_consume")
    artifacts: List[CalculatorArtifactResult] = Field(galias="reliquary_consume")

    @property
    def total(self) -> List[CalculatorConsumable]:
        artifacts = [i for a in self.artifacts for i in a.list]
        combined = self.character + self.weapon + self.talents + artifacts

        grouped: Dict[int, List[CalculatorConsumable]] = collections.defaultdict(list)
        for i in combined:
            grouped[i.id].append(i)

        total = [
            CalculatorConsumable(
                id=x[0].id,
                name=x[0].name,
                icon=x[0].icon,
                amount=sum(i.amount for i in x),
            )
            for x in grouped.values()
        ]

        return total


class CalculatorObject(NamedTuple):
    """An object required in the calculation of required materials"""

    id: int
    current: Optional[int]
    target: Optional[int]

    def _serialize(self, prefix: str = "") -> Dict[str, int]:
        """Serialize the object to a dict"""
        if self.id is None or self.current is None or self.target is None:
            raise ValueError("Cannot calculate with partial data")

        return {
            prefix + "id": int(self.id),
            prefix + "level_current": int(self.current),
            prefix + "level_target": int(self.target),
        }


class CalculatorBuilder:
    """A builder for the genshin impact enhancement calculator"""

    client: GenshinClient
    lang: Optional[str] = None

    character: Optional[CalculatorObject] = None
    weapon: Optional[CalculatorObject] = None
    artifacts: List[CalculatorObject]
    talents: List[CalculatorObject]

    character_element: Optional[int] = None
    full_artifact_set: bool = False
    include_talents_current_level: Optional[int] = None
    include_weapon: Optional[int] = None
    include_artifacts: Optional[Tuple[Optional[int], ...]] = None
    include_talents: Optional[Dict[str, Optional[int]]] = None

    def __init__(self, client: GenshinClient, lang: str = None) -> None:
        self.client = client
        self.lang = lang
        self.artifacts = []
        self.talents = []

    def set_lang(self, lang: str):
        """Set the language"""
        self.lang = lang

        return self

    def set_character(
        self,
        character: Union[int, BaseCharacter],
        current: int = None,
        target: int = None,
        *,
        element: int = None,
    ):
        """Set the character"""
        if isinstance(character, int):
            id = character
        else:
            id = character.id
            current = current or getattr(character, "level")

        self.character = CalculatorObject(id, current, target)
        self.character_element = element

        return self

    def set_weapon(self, id: int, current: int, target: int):
        """Set the weapon"""
        self.weapon = CalculatorObject(id, current, target)

        return self

    def add_artifact(self, id: int, current: int, target: int, *, full_set: bool = False):
        """Add an artifact"""
        self.artifacts.append(CalculatorObject(id, current, target))
        self.full_artifact_set = full_set

        return self

    def set_artifact_set(self, any_artifact_id: int, current: int, target: int):
        """Add an artifact set"""
        self.artifacts.append(CalculatorObject(any_artifact_id, current, target))
        self.full_artifact_set = True

        return self

    def add_talent(self, group_id: int, current: int, target: int):
        """Add a talent"""
        self.talents.append(CalculatorObject(group_id, current, target))

        return self

    def with_current_weapon(self, target: int):
        self.include_weapon = target

        return self

    def with_current_artifacts(
        self,
        target: int = None,
        *,
        flower: int = None,
        feather: int = None,
        sands: int = None,
        goblet: int = None,
        circlet: int = None,
    ):
        """Add all artifacts of the selected character"""
        if target:
            self.include_artifacts = (target,) * 5
        else:
            self.include_artifacts = (flower, feather, sands, goblet, circlet)

        return self

    def with_current_talents(
        self,
        target: int = None,
        current: int = None,
        *,
        attack: int = None,
        skill: int = None,
        burst: int = None,
    ):
        """Add all talents of the currently selected character"""
        if target:
            self.include_talents = {
                "attack": target,
                "skill": target,
                "burst": target,
            }
        else:
            self.include_talents = {
                "attack": attack,
                "skill": skill,
                "burst": burst,
            }

        self.include_talents_current_level = current

        return self

    async def build(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {}
        details: Optional[CalculatorCharacterDetails] = None

        if self.character_element:
            data["element_attr_id"] = self.character_element

        if self.character:
            data.update(self.character._serialize(prefix="avatar_"))

        if self.weapon:
            data["weapon"] = self.weapon._serialize()

        if self.artifacts:
            if self.full_artifact_set:
                if len(self.artifacts) != 1:
                    raise TypeError("Must have exactly 1 artifact to make a full set")

                original = self.artifacts[0]
                rest = await self.client.get_complete_artifact_set(original.id)
                self.artifacts += [
                    CalculatorObject(a.id, original.current, original.target) for a in rest
                ]

            data["reliquary_list"] = [i._serialize() for i in self.artifacts]

        if self.talents:
            data["skill_list"] = [i._serialize() for i in self.talents]

        if self.include_weapon:
            if self.character is None:
                raise TypeError("No character to get a weapon from")

            details = details or await self.client.get_character_details(self.character.id)

            x = CalculatorObject(details.weapon.id, details.weapon.level, self.include_weapon)
            data["weapon"] = x._serialize()

        if self.include_artifacts:
            if self.character is None:
                raise TypeError("No character to get artifacts from")

            details = details or await self.client.get_character_details(self.character.id)

            data["reliquary_list"] = []
            for artifact in details.artifacts:
                if target := self.include_artifacts[artifact.pos - 1]:
                    x = CalculatorObject(artifact.id, artifact.level, target)
                    data["reliquary_list"].append(x._serialize())

        if self.include_talents:
            if self.character is None:
                raise TypeError("No character to get talents from")

            current = self.include_talents_current_level or 0
            if current:
                talents = await self.client.get_character_talents(self.character.id)
            else:
                details = details or await self.client.get_character_details(self.character.id)
                talents = details.talents

            data["skill_list"] = []
            for talent in talents:
                if target := self.include_talents.get(talent.type):
                    x = CalculatorObject(talent.group_id, talent.level or current, target)
                    data["skill_list"].append(x._serialize())

        return data

    async def calculate(self) -> CalculatorResult:
        return await self.client._execute_calculator(self)

    def __await__(self) -> Generator[Any, None, CalculatorResult]:
        return self.calculate().__await__()



from __future__ import annotations

import re
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, validator

from .abyss import SpiralAbyssPair
from .activities import Activities
from .base import GenshinModel, PartialCharacter
from .character import Character

__all__ = [
    "Stats",
    "Offering",
    "Exploration",
    "TeapotRealm",
    "Teapot",
    "PartialUserStats",
    "UserStats",
    "FullUserStats",
]

# flake8: noqa: E222
class Stats(GenshinModel):
    """Overall user stats"""

    # This is such fucking bullshit, just why?
    # fmt: off
    achievements: int =       Field(galias="achievement_number",     mi18n="bbs/achievement_complete_count")
    days_active: int =        Field(galias="active_day_number",      mi18n="bbs/active_day")
    characters: int =         Field(galias="avatar_number",          mi18n="bbs/other_people_character")
    spiral_abyss: str =       Field(galias="spiral_abyss",           mi18n="bbs/unlock_portal")
    anemoculi: int =          Field(galias="anemoculus_number",      mi18n="bbs/wind_god")
    geoculi: int =            Field(galias="geoculus_number",        mi18n="bbs/rock_god")
    electroculi: int =        Field(galias="electroculus_number",    mi18n="bbs/electroculus_god")
    common_chests: int =      Field(galias="common_chest_number",    mi18n="bbs/general_treasure_box_count")
    exquisite_chests: int =   Field(galias="exquisite_chest_number", mi18n="bbs/delicacy_treasure_box_count")
    precious_chests: int =    Field(galias="precious_chest_number",  mi18n="bbs/rarity_treasure_box_count")
    luxurious_chests: int =   Field(galias="luxurious_chest_number", mi18n="bbs/magnificent_treasure_box_count")
    remarkable_chests: int =  Field(galias="magic_chest_number",     mi18n="bbs/magic_chest_number")
    unlocked_waypoints: int = Field(galias="way_point_number",       mi18n="bbs/unlock_portal")
    unlocked_domains: int =   Field(galias="domain_number",          mi18n="bbs/unlock_secret_area")
    # fmt: on

    def as_dict(self, lang: str = "en-us") -> Dict[str, Any]:
        """Helper function which turns fields into properly named ones"""
        return {
            self._get_mi18n(field, lang): getattr(self, field.name)
            for field in self.__fields__.values()
        }


class Offering(GenshinModel):
    """An offering"""

    name: str
    level: int


class Exploration(GenshinModel):
    """Exploration data"""

    id: int
    icon: str
    name: str
    type: str
    level: int
    explored: int = Field(galias="exploration_percentage")
    offerings: List[Offering]

    @property
    def percentage(self) -> float:
        """The percentage explored"""
        return self.explored / 10


class TeapotRealm(GenshinModel):
    """A specific teapot realm"""

    name: str
    icon: str

    @property
    def id(self) -> int:
        match = re.search(r"\d", self.icon)
        return int(match.group()) if match else 0


class Teapot(GenshinModel):
    """User's Serenitea Teapot"""

    realms: List[TeapotRealm]
    level: int
    visitors: int = Field(galias="visit_num")
    comfort: int = Field(galias="comfort_num")
    items: int = Field(galias="item_num")
    comfort_name: str = Field(galias="comfort_level_name")
    comfort_icon: str = Field(galias="comfort_level_icon")


class PartialUserStats(GenshinModel):
    """User stats with characters without equipment"""

    stats: Stats
    characters: List[PartialCharacter] = Field(galias="avatars")
    explorations: List[Exploration] = Field(galias="world_explorations")
    teapot: Optional[Teapot] = Field(galias="homes")

    @validator("teapot", pre=True)
    def __format_teapot(cls, v: Any) -> Optional[Dict[str, Any]]:
        if not v:
            return None
        if isinstance(v, dict):
            return v
        return {**v[0], "realms": v}


class UserStats(PartialUserStats):
    """User stats with characters with equipment"""

    characters: List[Character] = Field(galias="avatars")


class FullUserStats(UserStats):
    """User stats with all data a user can have"""

    abyss: SpiralAbyssPair
    activities: Activities



"""Pydantic models used by genshin.py"""
from . import *
from .abyss import *
from .activities import *
from .base import *
from .calculator import *
from .character import *
from .daily import *
from .diary import *
from .hoyolab import *
from .intermap import *
from .notes import *
from .stats import *
from .transaction import *
from .wish import *



import calendar
from datetime import datetime, timedelta, timezone
from typing import NamedTuple

from pydantic import Field

from .base import GenshinModel, Unique

__all__ = ["DailyRewardInfo", "DailyReward", "ClaimedDailyReward"]


class DailyRewardInfo(NamedTuple):
    """Information about the current daily reward status"""

    signed_in: bool
    claimed_rewards: int

    @property
    def missed_rewards(self) -> int:
        cn_timezone = timezone(timedelta(hours=8))
        now = datetime.now(cn_timezone)
        month_days = calendar.monthrange(now.year, now.month)[1]
        return month_days - self.claimed_rewards


class DailyReward(GenshinModel):
    """A claimable daily reward"""

    name: str
    amount: int = Field(galias="cnt")
    icon: str


class ClaimedDailyReward(GenshinModel, Unique):
    """A claimed daily reward"""

    id: int
    name: str
    amount: int = Field(galias="cnt")
    icon: str = Field(galias="img")
    time: datetime = Field(galias="created_at", timezone=8)
